
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "comm_interface_list.h"

#include "log.h"
#include "Searcher.h"
#include "evaluation.h"

#include <iostream>
#include <vector>


using namespace std;
using boost::lexical_cast;

namespace zochova
{

	namespace communicate
	{

		CommXboard::CommXboard() {
			cout << endl;
			cout.flush();
			logInputPrefix = "from xboard: ";
			logger = NULL;
			isShowThinking = true;
		}

		void CommXboard::sendOutput(string str) {
			if( logger != NULL ) {
				(*logger) << "to xboard:   " << str << "\n";
			}
			cout << "---------------------------------------------------------------" << endl;
			cout << str << endl;
			cout.flush();
		}

		bool CommXboard::newInput(string command) {
			Move m;
			if( command == "xboard" ) {
				sendOutput("\n");
			}
			else if( command.substr(0, 8) == "protover" ) {
				sendOutput( (new string())
					->append("feature ping=1 usermove=0 san=0 ")
					.append("myname=\"")
					.append(ZOCHOVA_NAME)
					.append(" ")
					.append(ZOCHOVA_VERSION)
					.append("\" done=1") );
			}
			else if( command == "new" ) {
				eventNewGame();
				eventSetEngineBlack();
			}
			else if( command == "resign" ) {
				eventResign();
			}
			else if( command == "quit" ) {
//				eventQuit();
				return false;
			}
			else if( command == "black" ) {
				eventSetEngineBlack();
			}
			else if( command == "white" ) {
				eventSetEngineWhite();
			}
			else if( command == "post" ) {
				isShowThinking = true;
			}
			else if( command == "nopost" ) {
				isShowThinking = false;
			}
			else if( command == "hard" ) {
				eventSetPonder();
			}
			else if( command == "easy" ) {
				eventUnSetPonder();
			}
			else if( command == "force" ) {
				eventForce();
			}
			else if( command == "?" ) {
				eventMoveNow();
			}
			else if( command.substr(0, 4) == "time" ) {
				try {
					eventSetEngineTime(lexical_cast<int>(command.substr(5)));
				}
				catch( ... ) { }
			}
			else if( command.substr(0, 4) == "otim" ) {
				try {
					eventSetUserTime(lexical_cast<int>(command.substr(5)));
				}
				catch( ... ) { }
			}
			else if( command.substr(0, 4) == "ping" ) {
				sendOutput(command.replace(1, 1, "o"));
			}
			else if( command.substr(0, 5) == "level" ) {
				vector<string> parts;
				stringSplit(command, " ", parts);
				if( parts.size() == 4 ) {
					try {
						eventSetTimeControl(
							lexical_cast<int>(parts.at(1)),
							lexical_cast<int>(parts.at(2)) * 60*100,
							lexical_cast<int>(parts.at(3)) * 100 );
					}
					catch( ... ) { }
				}
			}
			else if( readMove(m, command) ) {
				eventMakeMove(m);
			}
			processed = true;
			return true;
		} // CommXboard::newInput

		void CommXboard::makeMove(const Move &m) {
			if( !isUsersTurn ) {
				sendOutput("move "+m.print());
			}
	//		eventNeedCurrentState(&currentState);
	//		cout << currentState.print();
		}

		void CommXboard::newThinkingLine(const engine::search::Result& result) {
			if( isShowThinking ) {
				sendOutput(
					lexical_cast<string>(result.getDepth())+
					" "+lexical_cast<string>(((int)result.getValue()) * 100 / engine::evaluation::PAWN_VALUE)+
					" "+lexical_cast<string>(result.getTime())+
					" "+lexical_cast<string>(result.getNodeCount())+
					" "+printMoveList(result.getMoves()));
			}
		}

		void CommXboard::resign() {
			sendOutput("resign");
			setUsersTurn();
		}
		
	} // communicate

} // zochova
