
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "TranspositionTable.h"
#include "Random64.h"



#include "Searcher.h"



#define TTLIST


namespace zochova
{
	namespace zobrist
	{

		uint64 delta(uint64 key,
							  int s,
							  int d,
							  int pi,
							  int cap_pi,
							  Side next,
							  int pr_pi)
		{
			uint64 r;
			if( pr_pi >= noPieceIndex ) {
				pr_pi = pi;
			}
			if( next == WHITE ) {
				s = 8*(7-s/8)+s%8;
				d = 8*(7-d/8)+d%8;
			}
			r = key ^ RandomPiece[pi*64 + s] ^ RandomPiece[pr_pi*64 + d];
			if( cap_pi < noPieceIndex ) {
				r ^= RandomPiece[cap_pi*64 + d];
			}
			return r;
		}

		uint64 delta_p(uint64 key,
							  int s,
							  int d,
							  int pi,
							  int cap_pi,
							  int pr_pi,
							  Side next)
		{
			uint64 r;
			if( next == WHITE ) {
				s = 8*(7-s/8)+s%8;
				d = 8*(7-d/8)+d%8;
			}
			r = key ^ RandomPiece[pi*64 + s] ^ RandomPiece[pr_pi*64 + d];
			if( cap_pi < noPieceIndex ) {
				r ^= RandomPiece[cap_pi*64 + d];
			}
			return r;
		}

		TransPositionEntry::TransPositionEntry(unsigned __int32 value) {
			fromInt(value);
		}

		void TransPositionEntry::fromInt(unsigned __int32 value) {
			source = value&63;
			dest = (value/64)&63;
			searchValue = (((value/(64*64)&4095)))-2000;
			isUpperBound = ((value&(64*64*4096*2)) != 0);
			isLowerBound = ((value&(64*64*4096*4)) != 0);
			nullMoveVerify = ((value&(64*64*4096)) != 0);
			ply = value/(64*64*4096*8);
		}

		TranspositionTable::TranspositionTable() {
			clear();
		}

		bool TranspositionTable::lookup(uint64 hashKey64, TransPositionEntry &entry) {
			#ifdef TTLIST

			unsigned int hashUpper = ((hashKey64&HASH_UPPER)>>32);
			unsigned int hashLower = hashKey64&HASH_LOWER;
			int i;
			i = 0;
			while( table[hashLower+i].hash_upper != hashUpper ) {
				if( table[hashLower+i].hash_upper == 0 ) {
					return false;
				}
				++i;
				if( i == 8 ) {
					return false;
				}
			}
			entry.fromInt(table[hashLower+i].data);
			return true;


			#else

			if( table[hashKey64&HASH_LOWER].hash_upper == 0) {
				return false;
			}
			if( table[hashKey64&HASH_LOWER].hash_upper == ((hashKey64&HASH_UPPER)>>32) ) {
				entry.fromInt(table[hashKey64&HASH_LOWER].data);
				return true;
			}
			else {
				return false;
			}

			#endif
			
		}

		void TranspositionTable::insert(uint64 hashKey64,
								   int source,
								   int dest,
								   int value,
								   bool isUpperBound,
								   bool isLowerBound,
								   unsigned int ply,
								   bool nullMoveVerify)
		{
/*			if( table[hashKey64&HASH_LOWER].hash_upper ) {
				engine::search::Searcher::CoUnTeR++;
			}*/
			#ifdef TTLIST

			unsigned int hashUpper = ((hashKey64&HASH_UPPER)>>32);
			unsigned int hashLower = hashKey64&HASH_LOWER;
			int i;
			i = 0;
			while( table[hashLower+i].hash_upper != hashUpper )
			{
				if( table[hashLower+i].hash_upper == 0 ) {
					table[hashLower+i].hash_upper = hashUpper;
					break;
				}
				++i;
				if( i == 8 ) {
	//				engine::search::Searcher::CoUnTeR++;
					i = 0;
					table[hashLower].hash_upper = hashUpper;
					break;
				}
			}
			table[hashLower+i].data = 
				( (ply)*(64*64*4096*8) )
				+ ( isLowerBound ? (64*64*4096*4) : 0 )
				+ ( isUpperBound ? (64*64*4096*2) : 0 )
				+ ( nullMoveVerify ? (64*64*4096) : 0 )
				+ (((unsigned int)(value+2000))*64*64)
				+ dest*64
				+ source;
/*			if( table[hashLower+i].data & (64*64*4096) ) {
				i++;
			}
*/			#else

			table[hashKey64&HASH_LOWER].hash_upper = ((hashKey64&HASH_UPPER)>>32);
			table[hashKey64&HASH_LOWER].data =
				( (ply)*(64*64*4096*8) )
				+ ( isLowerBound ? (64*64*4096*4) : 0 )
				+ ( isUpperBound ? (64*64*4096*2) : 0 )
				+ (((unsigned int)(value+2000))*64*64)
				+ dest*64
				+ source;

			#endif
		}

		void TranspositionTable::remove(uint64 hashKey64) {
			table[hashKey64&HASH_LOWER].hash_upper = 0;
		}

		void TranspositionTable::clear() {
			memset(table, 0, sizeof(table));
		}

		uint64 calculate(const State &state) {
			int pos = 64;
			uint64 z = 0;
			while( pos ) {
				pos--;
				switch( state[(BoardPosition)pos] ) {
					case WHITE_PAWN:
						z ^= Random64[whitePawnIndex*64 + pos];
						break;
					case WHITE_KNIGHT:
						z ^= Random64[whiteKnightIndex*64 + pos];
						break;
					case WHITE_ROOK:
						z ^= Random64[whiteRookIndex*64 + pos];
						break;
					case WHITE_BISHOP:
						z ^= Random64[whiteBishopIndex*64 + pos];
						break;
					case WHITE_QUEEN:
						z ^= Random64[whiteQueenIndex*64 + pos];
						break;
					case WHITE_KING:
						z ^= Random64[whiteKingIndex*64 + pos];
						break;
					case BLACK_PAWN:
						z ^= Random64[blackPawnIndex*64 + pos];
						break;
					case BLACK_KNIGHT:
						z ^= Random64[blackKnightIndex*64 + pos];
						break;
					case BLACK_ROOK:
						z ^= Random64[blackRookIndex*64 + pos];
						break;
					case BLACK_BISHOP:
						z ^= Random64[blackBishopIndex*64 + pos];
						break;
					case BLACK_QUEEN:
						z ^= Random64[blackQueenIndex*64 + pos];
						break;
					case BLACK_KING:
						z ^= Random64[blackKingIndex*64 + pos];
						break;
				}
/*				if( state[(BoardPosition)pos] != EMPTY ) {
					
					z ^= Random64[(LS1B_INDEX(SquareToPiece(state[(BoardPosition)pos]))-2)*64+pos];
				}*/
			}
			/*if( st->next == BLACK ) {
				key ^= Random64[11*64+63+1];
			}*/
			if( state.canBlackCastleLeft() ) {
				z ^= blackCastleLeft;
			}
			if( state.canBlackCastleRight() ) {
				z ^= blackCastleRight;
			}
			if( state.canWhiteCastleLeft() ) {
				z ^= whiteCastleLeft;
			}
			if( state.canWhiteCastleRight() ) {
				z ^= whiteCastleRight;
			}
			return z;
		}

	} // zobrist
} // zochova
