
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "search_flags.h"
#include "Searcher.h"
#include "TranspositionTable.h"
#include "evaluation.h"
#include "Thinker.h"
#if ENGINE_LOG_LEVEL > 1
#	include "log.h"
#endif


namespace zochova
{
	namespace engine
	{
		namespace search
		{
			int Searcher::_search(Data* data,
								  int dynamicValue,
								  uint64 &nodeCount,
								  bool nullVerify,
								  unsigned int ply,
								  bool inLateMove)
			{
			#	define data_BB_ARGUMENTS_CALL \
					data->all_p, \
					data->is_my, \
					data->is_op, \
					data->is_pawn, \
					data->is_r, \
					data->is_b, \
					data->k_pos

				using namespace zobrist;
				using namespace evaluation;

				uint64 attacks;
				uint64* m;
				uint64 transposMove;
				bool isUpperBound, isLowerBound;
				uint64 bestMove;
				bestMove = 0;
				#ifdef EVAL_STATIC

				int staticValue;
				bool haveStaticValue;
				haveStaticValue = false;
				staticValue = 0;

				#endif
				
				int newDynamicValue = 0;


				assert( data->transpositionTable != 0 );
				data->nodes++;
				++nodeCount;
/*				if( (nodeCount == 3532785) && (ply == 68) ) {
					int sdfg;
sdfg = 3;
++sdfg;
				}
*/
// uint64 tnc;
// tnc = -1;
// tnc = nodeCount;

				if( (nodeCount % 1024) == 0 ) {
					//
					//  Time control 
					if( (data->timer != 0) && data->timer->timeOut() ) {
						return -evaluation::SEARCH_HALT_VALUE;
					}
				}
				#ifdef PV_STORE
				data->bestLine[0] = 0;
				#endif
				Data* d;
				d = data-1;
				while( d->all_p != 0 ) {
					if( d->hashKey == data->hashKey ) {
						return 0;
					}
					--d;
				}
				#ifdef QSEARCH
				if( ply == 0 ) {
				#else
				if( ply <= (PLY_FULL_SEARCH_MIN) ) {
				#endif
					#ifdef EVAL_STATIC
					return evaluation::staticValue(data_BB_ARGUMENTS_CALL);
					#else
					return dynamicValue;
					#endif
				}
/*				if( ply <= 30 ) {
					isThreat = false;
				}
*/				transposMove = 0;

				bool checked;

				bool ttrecheking;
				ttrecheking = false;
				bool allChildren;

				allChildren = (ply > PLY_FULL_SEARCH_MIN);
				/*if( allChildren ) {
					++CoUnTeR;
				}
*/
				if( !allChildren ) {

					assert( data->transpositionTable != 0 );
					#ifdef EVAL_STATIC

					if( !haveStaticValue ) {
						staticValue = evaluation::staticValue(data_BB_ARGUMENTS_CALL);
					}
					if( staticValue > data->alpha ) {
						data->alpha = staticValue;
						if( staticValue >= data->beta ) {
							assert( data->transpositionTable != 0 );
							assert( staticValue < evaluation::INFINITE_VALUE );
							assert( staticValue > -evaluation::INFINITE_VALUE );
							return staticValue;
						}
					}

					#else

/*					if( (data->alpha >= dynamicValue + QUEEN_VALUE + PAWN_VALUE) && (ply <= PLY_FULL_SEARCH_MIN-4) ) {
						return dynamicValue;
					}
*/
					if( bitboardmoves::kingChecked(
						data->all_p,
						data->is_op,
						data->is_my,
						data->is_pawn,
						data->is_r,
						data->is_b,
						data->k_pos,
						true) ) 
					{
						return -SEARCH_ILLEGAL_MOVE_VALUE;
					}
					if( dynamicValue > data->alpha ) {
						data->alpha = dynamicValue;
						if( dynamicValue >= data->beta ) {
							return dynamicValue;
						}
					}


					#endif
				}


				RECHEK_TT:

				#ifdef USE_TP

				if(
				#ifndef USE_QTP
					( ply > PLY_FULL_SEARCH_MIN ) 
					&&
				#endif

					data->transpositionTable->lookup(data->hashKey,
													 data->transponsEntry) )
	//			if( false )
				{
							

					transposMove = Move::toBitBoard(
						data->transponsEntry.source,
						data->transponsEntry.dest,
						data->is_my,
						data->is_op,
						data->k_pos,
						data->is_pawn);
					if( (data->transponsEntry.ply + PLY_FULL_SEARCH_MIN-16 >= ply)
						&& ( (!nullVerify) 
						     || data->transponsEntry.nullMoveVerify
							 || (ply <= PLY_FULL_SEARCH_MIN+1)
							 || (data->transponsEntry.ply + PLY_FULL_SEARCH_MIN-16 > ply+1)) )	
					{
						assert( data->transponsEntry.searchValue < evaluation::INFINITE_VALUE );
						assert( data->transponsEntry.searchValue > -evaluation::INFINITE_VALUE );
						if( data->transponsEntry.isLowerBound ) {
							if( data->beta <= data->transponsEntry.searchValue )  {
								#ifdef PV_STORE
								data->bestLine[0] = transposMove;
								data->bestLine[1] = 0;
								#endif

								return data->transponsEntry.searchValue;
							}
							if( data->alpha < data->transponsEntry.searchValue ) {
								data->alpha = data->transponsEntry.searchValue;
							}
						}
						else {
							if( data->transponsEntry.isUpperBound ) {
								if( data->alpha >= data->transponsEntry.searchValue )  {
									#ifdef PV_STORE
									data->bestLine[0] = transposMove;
									data->bestLine[1] = 0;
									#endif

									return data->alpha;
								}
							}
							else {
								#ifdef PV_STORE
								data->bestLine[0] = transposMove;
								data->bestLine[1] = 0;
								#endif
								return data->transponsEntry.searchValue;
							}
						}
					}
				}
				#endif
/*
				if( (!ttrecheking) && (transposMove == 0) && (ply > PLY_FULL_SEARCH_MIN+6) ) {
					int t;
					t = _search(data, dynamicValue, nodeCount, nullVerify, PLY_FULL_SEARCH_MIN+2, inLateMove);
					if( t == SEARCH_HALT_VALUE ) {
						return -SEARCH_HALT_VALUE;
					}
					if( t == SEARCH_ILLEGAL_MOVE_VALUE ) {
						return -SEARCH_ILLEGAL_MOVE_VALUE;
					}
					++CoUnTeR;
					ttrecheking = true;
					goto RECHEK_TT;
				}
*/
				checked = bitboardmoves::kingChecked(data_BB_ARGUMENTS_CALL);

				#ifdef USE_SEARCH_EXTENSIONS

				if( data->canExtend ) {
					if( checked ) {
						data->extendPly += 2;
					}
					if( data->isPVNode ) {
						data->extendPly += 1;
					}
					if( (ply == PLY_FULL_SEARCH_MIN+1) ) {
						if( data->extendPly > 6 ) {
							data->extendPly = 6;
						}
						if( data->extendPly > 5 ) {
							++CoUnTeR;
						}
						++CoUnTeR2;
						ply = ply + (data->extendPly/3);
						(data+1)->extendPly = 0;
						(data+1)->canExtend = false;
					}
					else {
						(data+1)->extendPly = data->extendPly;
						(data+1)->canExtend = true;
					}
				}
				else {
					(data+1)->canExtend = false;
				}
				#endif

/*
				#ifdef USE_MOVE_ORDERING_2

				uint64 att_p, att_b, att_r, att_k, att_disc;

				if( bitboardmoves::kingCheckSquares(
					data_BB_ARGUMENTS_CALL, att_p, att_b, att_r, att_k, att_disc) )
				{
					return -SEARCH_ILLEGAL_MOVE_VALUE;
				}

				#endif
*/
		//		bool tval_;
		//		tval_ = false;
/*
				if( (!checked) && (ply <= PLY_FULL_SEARCH_MIN) ) {
					tval = evaluation::staticValue(data_BB_ARGUMENTS_CALL);
					tval_ = true;
					if( tval <= (data->alpha - evaluation::BISHOP_VALUE) ) {
						return data->alpha;
					}
				}
*/
				Data* data1;


				#ifdef USE_NULLMOVE_PRUNE

				bool failHigh;
				failHigh = false;
				int originalPly;
				if( (!checked) //&& false
					&& (ply > PLY_FULL_SEARCH_MIN+3) 
				//	&& (ply > PLY_FULL_SEARCH_MIN+NULLMOVE_PRUNE_PLY)
				//	&& ( (ply > PLY_FULL_SEARCH_MIN + NULLMOVE_PRUNE_PLY) || ((!nullVerify) && (!inLateMove)) )
				//	&& (ply <= PLY_FULL_SEARCH_MIN)
				//	&& (ply > 16)
				//	&& nullVerify 
					&& (!data->isPVNode)
					&& ((data-1)->hashKey != 0)
				//	&& (!inLateMove)
					&& (POPCOUNT(data->is_my & (~data->is_pawn)) > 1) 
					
					)
				{
					#ifdef EVAL_STATIC

					if( !haveStaticValue ) {
						staticValue = evaluation::staticValue(data_BB_ARGUMENTS_CALL);
						haveStaticValue = true;
					}
					if( staticValue >= data->beta  + QUEEN_VALUE ) {

					#else

					if( dynamicValue >= data->beta + QUEEN_VALUE ) {

					#endif
					//if( true ) {
						data1 = data+1;
						data1->all_p = FLIPVERTICAL(data->all_p);
						data1->is_my = FLIPVERTICAL(data->is_op);
						data1->is_op = FLIPVERTICAL(data->is_my);
						data1->is_pawn = FLIPVERTICAL(data->is_pawn);
						data1->is_r = FLIPVERTICAL(data->is_r);
						data1->is_b = FLIPVERTICAL(data->is_b);
						data1->k_pos = FLIPVERTICAL(data->k_pos);

				//		#ifndef USE_MOVE_ORDERING_2

						if( bitboardmoves::kingChecked(
							data1->all_p,
							data1->is_my,
							data1->is_op,
							data1->is_pawn,
							data1->is_r,
							data1->is_b,
							data1->k_pos) )
						{
							return -SEARCH_ILLEGAL_MOVE_VALUE;
						}

				//		#endif

						data1->castle_info = state_castleinfo::FLIP(data->castle_info);
						data1->ep = 0;
						data1->hashKey = data->hashKey;
						data->hashKey = 0;
						data1->alpha = -data->beta;
						data1->beta = -data->beta+1;
						data1->isPVNode = false;
						data1->isAllNode = false;
						data1->isCutNode = true;
						#ifdef USE_COUNTER_MOVES

						data1->prevMove = 0;

						#endif
						int v;

						v = -_search(data1,
									 -dynamicValue,
									 nodeCount,
									 nullVerify,
									 ply - NULLMOVE_PRUNE_PLY -1,
									 true);

						if( v == evaluation::SEARCH_HALT_VALUE ) {
							return -evaluation::SEARCH_HALT_VALUE;
						}
						assert( v < evaluation::INFINITE_VALUE );
						assert( v > -evaluation::INFINITE_VALUE );
						data->hashKey = data1->hashKey;
						if( v >= data->beta ) {
							#ifdef USE_VERIFYED_NULLMOVE
							if( nullVerify ) {
								failHigh = true;
								nullVerify = false;
								originalPly = ply;
								if( dynamicValue > data->beta + QUEEN_VALUE + PAWN_VALUE ) {
									ply -= 2;
								}
								else {
									--ply;
								}
							}
							else {
							//	++CoUnTeR;
								return v;
							}
							#else
							return v;
							#endif
						}
					}
				}
				#endif

				assert( data->transpositionTable != 0 );
				uint64 *listEnd;
				listEnd = (uint64*)1;

				if( allChildren ) {
					attacks = bitboardmoves::gen(
									data_BB_ARGUMENTS_CALL,
									#ifdef USE_MOVE_ORDERING_2
									data->MoveList+5,
									#else
									data->MoveList,
									#endif
									&listEnd);
					if( (*(data->MoveList+5)) == 0 ) {
						return 0;
					}
				}
				else {
					
					attacks = bitboardmoves::gen(
									data_BB_ARGUMENTS_CALL,
									#ifdef USE_MOVE_ORDERING_2
									data->MoveList+5,
									#else
									data->MoveList,
									#endif
									&listEnd,
									true);
					#ifdef USE_MOVE_ORDERING_2
					if( (*(data->MoveList+5)) == 0 ) {
					#else
					if( (*(data->MoveList)) == 0 ) {
					#endif
						#ifdef EVAL_STATIC

						if( !haveStaticValue ) {
							staticValue = evaluation::staticValue(data_BB_ARGUMENTS_CALL);
							haveStaticValue = true;
						}
						return staticValue;

						#else

						return dynamicValue;

						#endif

					}
				}

				assert( data->transpositionTable != 0 );
				if( attacks&data->k_pos ) {
					assert( data->transpositionTable != 0 );
					return -evaluation::SEARCH_ILLEGAL_MOVE_VALUE;
				}


/*
				if( !isThreat ) {
					if( bitboardmoves::kingChecked(data_BB_ARGUMENTS_CALL) ) {
						isThreat = true;
					}
				}
*/

				if( ply > PLY_FULL_SEARCH_MIN )

				if( data->ep != 0 ) {
					bitboardmoves::genEnp(data->is_my,
										  data->is_pawn,
										  data->ep,
										  listEnd);
				}
				assert( data->transpositionTable != 0 );


				//if( (ply < PLY_CHECK_SEARCH_MAX) || !checked ) {

				if( (ply > PLY_FULL_SEARCH_MIN) && !checked ) {
					if( state_castleinfo::GET_MY_CASTLE_LEFT(data->castle_info) ) {
						bitboardmoves::genCastleLeft(data_BB_ARGUMENTS_CALL, listEnd);
					}
					if( state_castleinfo::GET_MY_CASTLE_RIGHT(data->castle_info) ) {
						bitboardmoves::genCastleRight(data_BB_ARGUMENTS_CALL, listEnd);
					}
					assert( data->transpositionTable != 0 );
				}
				m = data->MoveList;

				#ifdef USE_MOVE_ORDERING_2

				if( !checked ) {
					uint64 k0, k1, k2;
					k0 = 0;

					#ifdef USE_COUNTER_MOVES

					if( data->prevMove == 0 ) {
						k0 = data->nullCounterMove;
					}
					else {
						if( data->prevMove == CL_SQUARES ) {
							k0 = data->clCounterMove;
						}
						else {
							if( data->prevMove == CR_SQUARES ) {
								k0 = data->crCounterMove;
							}
							else {
								if( (data->prevMove&M_PROMOTE_ANY) == M_PROMOTE_ANY ) {
									data->prevMove = data->prevMove^M_PROMOTE_ANY;
								}
								if( (data->prevMove&M_ENP) == M_ENP ) {
									data->prevMove = data->prevMove^M_ENP;
								}
								int ps, pd;
								ps = LS1B_INDEX(((data-1)->is_my)&data->prevMove);
								pd = LS1B_INDEX((~((data-1)->is_my))&data->prevMove);
								k0 = data->counterMoveTable[ps][pd];
							}
						}
					}

					#endif

					k1 = data->killer;
					k2 = data->killer1;

					moveOrder(
						data_BB_ARGUMENTS_CALL,
						data->MoveList,
						transposMove,
						k0,
						k1,
						k2);
						
						
		/*
					if( 
						
						
						!moveOrder(
						data_BB_ARGUMENTS_CALL,
						data->MoveList,
						data->MoveList2,
						transposMove,
						data->killer,
						data->killer1,
						data->killer2,
						data->is_op&(~data->is_pawn),
						data->is_op&data->is_pawn,
						att_p, att_b, att_r, att_k, att_disc,
						(ply > PLY_FULL_SEARCH_MIN) ? true : false,
						(ply > PLY_FULL_SEARCH_MIN) ? true : false
						//(ply > PLY_CHECK_SEARCH_MIN) ? true : false
						
						
						) 
						
						
						
						
						
						
						
						
						
						)
					{
						#ifdef EVAL_STATIC

						if( !haveStaticValue ) {
							staticValue = evaluation::staticValue(data_BB_ARGUMENTS_CALL);
							haveStaticValue = true;
						}
						return staticValue;

						#else

						return dynamicValue;

						#endif
					}
					*/
					m = data->MoveList;
					//allChildren = (ply > PLY_FULL_SEARCH_MIN);
				}
				else {
					m = data->MoveList+5;
					//allChildren = true;
				}
					

				#endif

				//int treshold;
/*				if( isThreat || (ply > PLY_FULL_SEARCH_MIN) ) {
					//treshold =  -evaluation::INFINITE_MOVE_VALUE;
				}
				else {
					//treshold = (96/PLY_FULL_SEARCH_MIN) * (PLY_FULL_SEARCH_MIN - ply)+222;
				}
*/
			/*	m = moveOrder(data->MoveList,
							  data_BB_ARGUMENTS_CALL,
						//	  data->killerMoves,
						//	  data->moveTreshold,
							  treshold,
							  transposMove,
							  data->moveValues,
							  data->threats);
					//tnc = m - data->MoveList;
				/*	uint64 * tx;
					tx = m;
					tnc = 0;
					while( *tx ) {
						++tnc;
						++tx;
					}
				*/	
				assert( data->transpositionTable != 0 );

				#pragma region MOVE_ORDERING_1
				#ifndef USE_MOVE_ORDERING_2

				if( (*m) == 0 ) {
					assert( data->transpositionTable != 0 );
					#ifdef EVAL_STATIC

					if( !haveStaticValue ) {
						staticValue = evaluation::staticValue(data_BB_ARGUMENTS_CALL);
					}
					assert( staticValue < evaluation::INFINITE_VALUE );
					assert( staticValue > -evaluation::INFINITE_VALUE );

					return staticValue;

					#else

					return dynamicValue;

					#endif
				}
				uint64 * first;
				first = m;
				#ifdef USE_TP

				if( transposMove != 0 ) {
					++m;
					while( (*m) != 0 ) {
						if( (*m) == transposMove ) {
							(*m) = (*(data->MoveList));
							(*(data->MoveList)) = transposMove;
							++first;
							break;
						}
						++m;
					}
				}
				#endif

				#if KILLER_MOVE_COUNT > 0

				if( data->killer != 0 ) {
					if( data->killer != transposMove ) {
						m = first+1;
						while( (*m) != 0 ) {
							if( (*m) == data->killer ) {
								(*m) = (*first);
								(*first) = data->killer;
								++first;
								//++CoUnTeR;
								break;
							}
							++m;
						}
					}
					else {
						//++CoUnTeR;
					}
				}
				#endif

				#if KILLER_MOVE_COUNT > 1

				if( data->killer1 != 0 ) {
					if( data->killer1 != transposMove ) {
						m = first+1;
						while( (*m) != 0 ) {
							if( (*m) == data->killer1 ) {
								(*m) = (*first);
								(*first) = data->killer1;
								++first;
								//++CoUnTeR;
								break;
							}
							++m;
						}
					}
					//else {
					//	++CoUnTeR;
					//}
				}
				#endif

				#if KILLER_MOVE_COUNT > 2

				if( data->killer2 != 0 ) {
					if( data->killer2 != transposMove ) {
						m = first+1;
						while( (*m) != 0 ) {
							if( (*m) == data->killer2 ) {
								(*m) = (*first);
								(*first) = data->killer2;
								++first;
								//++CoUnTeR;
								break;
							}
							++m;
						}
					}
					//else {
					//	++CoUnTeR;
					//}
				}
				#endif

				uint64 * pbestMove;
				pbestMove = 0;
				if( ply > PLY_FULL_SEARCH_MIN ) {
					int bestMoveValue;
					bestMoveValue = 0;
					for( m = first; (*m) != 0 ; ++m ) {
						if( ((*m) & M_PROMOTE_ANY) == M_PROMOTE_ANY ) {
							pbestMove = m;
							break;
						}
						if( ((*m) == CR_SQUARES) || ((*m) == CL_SQUARES) ) {
							continue;
						}
						if( ((*m) & M_ENP) == M_ENP ) {
							if( bestMoveValue < 1 ) {
								bestMoveValue = 1;
								pbestMove = m;
								continue;
							}
						}
						if( ((*m) & data->is_op) == 0 ) {
							continue;
						}
						if( ((*m) & data->is_op & data->is_b & data->is_r) ) {
							if( (*m) & data->is_my & data->is_b & data->is_r ) {
								if( bestMoveValue < 8 ) {
									bestMoveValue = 8;
									pbestMove = m;
									continue;
								}
							}
							else {
								bestMoveValue = 10;
								pbestMove = m;
								continue;
							}
						}
						if( (*m) & data->is_op & (~data->is_pawn) ) {
							if( (*m) & data->is_my & data->is_pawn ) {
								if( bestMoveValue < 7 ) {
									bestMoveValue = 7;
									pbestMove = m;
									continue;
								}
							}
							else {
								if( bestMoveValue < 6 ) {
									bestMoveValue = 6;
									pbestMove = m;
									continue;
								}
							}
						}
						if( bestMoveValue < 1 ) {
							bestMoveValue = 1;
							pbestMove = m;
						}
					}
					if( pbestMove != 0 ) {
						uint64 t;
						t = (*pbestMove);
						(*pbestMove) = (*first);
						(*first) = t;
						++first;
					}
				}

				m = data->MoveList;
				#endif
				#pragma endregion

				/*
				if( !allChildren ) {
					assert( data->transpositionTable != 0 );
					#ifdef EVAL_STATIC

					if( !haveStaticValue ) {
						staticValue = evaluation::staticValue(data_BB_ARGUMENTS_CALL);
					}
					if( staticValue > data->alpha ) {
						data->alpha = staticValue;
						if( staticValue >= data->beta ) {
							assert( data->transpositionTable != 0 );
							assert( staticValue < evaluation::INFINITE_VALUE );
							assert( staticValue > -evaluation::INFINITE_VALUE );
							return staticValue;
						}
					}

					#else

					if( dynamicValue > data->alpha ) {
						data->alpha = dynamicValue;
						if( dynamicValue >= data->beta ) {
							++CoUnTeR;
							return dynamicValue;
						}
					}

					#endif
				}
			*/
				int originalAlpha = data->alpha;
				uint64 * originalM = m;

				int bestReturned;

				bool underReSearch;
				underReSearch = false;
				RE_SEARCH:


				assert( data->transpositionTable != 0 );

				data1 = data+1;

				isLowerBound = false;
				isUpperBound = true;

				bestReturned = -evaluation::INFINITE_VALUE;

				int childrenSearched;
				childrenSearched = 0;

				if( data->isPVNode ) {
					data1->isPVNode = true;
					data1->isAllNode = false;
					data1->isCutNode = false;
				}
				else {
					if( data->isCutNode ) {
						data1->isPVNode = false;
						data1->isAllNode = true;
						data1->isCutNode = false;
					}
					else {
						data1->isPVNode = false;
						data1->isAllNode = false;
						data1->isCutNode = true;
					}
				}

				do {
					#ifdef USE_MOVE_ORDERING_2

					if( (*m) == goto_next_move ) {
						++m;
						continue;
					}

					if( (m > data->MoveList+4) && (!underReSearch) ) {
						nextMove(data_BB_ARGUMENTS_CALL, m);
					}

					#endif

					#ifdef USE_COUNTER_MOVES

					data1->prevMove = (*m);

					#endif

					#ifdef USE_SEARCH_EXTENSIONS

					if( data->canExtend ) {
						(data+1)->extendPly = data->extendPly;
						if( (*m) & data->is_op ) {
							(data+1)->extendPly += 1;
						}

					}

					#endif

					assert( (*m) & (data->is_my) );
					assert( data->transpositionTable != 0 );
					int sourcePiece;
					int destPiece;
					
					assert( data->transpositionTable != 0 );

					data1->all_p = FLIPVERTICAL(data->all_p);
					data1->is_my = FLIPVERTICAL(data->is_op);
					data1->is_op = FLIPVERTICAL(data->is_my);
					data1->is_pawn = FLIPVERTICAL(data->is_pawn);
					data1->is_r = FLIPVERTICAL(data->is_r);
					data1->is_b = FLIPVERTICAL(data->is_b);
					data1->k_pos = FLIPVERTICAL(data->k_pos);

					state_castleinfo::CastleInfo tCastleInfo;
					tCastleInfo = data->castle_info;

					assert( (bitboardmoves::king_moves[LS1B_INDEX(data->is_my&data->k_pos)] & data->is_op&data->k_pos) == 0);

					bitboardmoves::makeMoveFlipped(
						data1->all_p,
						data1->is_op,
						data1->is_my,
						data1->is_pawn,
						data1->is_r,
						data1->is_b,
						data1->k_pos,
						*m,
						sourcePiece,
						destPiece,
						tCastleInfo,
						data->next,
						data1->ep);

					assert( (bitboardmoves::king_moves[LS1B_INDEX(data->is_my&data->k_pos)] & data->is_op&data->k_pos) == 0);

					assert( POPCOUNT(data1->is_my) < 17 );
					assert( POPCOUNT(data1->is_op) < 17 );
					data1->castle_info = state_castleinfo::FLIP(tCastleInfo);
					data1->hashKey = data->hashKey;
					assert( data->transpositionTable != 0 );

					if( (*m) == CR_SQUARES ) {
						#ifndef EVAL_STATIC

						newDynamicValue = dynamicValue + CASTLED_KING_VALUE;

						#endif
						if( data->next == WHITE ) {
							data1->hashKey ^= zobrist::xorWhiteCastleRight;
							if( state_castleinfo::GET_MY_CASTLE_LEFT(data->castle_info) ) {
								data1->hashKey ^= zobrist::whiteCastleLeft;
							}
						}
						else {
							data1->hashKey ^= zobrist::xorBlackCastleRight;
							if( state_castleinfo::GET_MY_CASTLE_LEFT(data->castle_info) ) {
								data1->hashKey ^= zobrist::blackCastleLeft;
							}
						}
					}
					else {
						if( (*m) == CL_SQUARES ) {
							#ifndef EVAL_STATIC

							newDynamicValue = dynamicValue + CASTLED_KING_VALUE;

							#endif
							if( data->next == WHITE ) {
								data1->hashKey ^= zobrist::xorWhiteCastleLeft;
								if( state_castleinfo::GET_MY_CASTLE_RIGHT(data->castle_info) ) {
									data1->hashKey ^= zobrist::whiteCastleRight;
								}
							}
							else {
								data1->hashKey ^= zobrist::xorBlackCastleLeft;
								if( state_castleinfo::GET_MY_CASTLE_RIGHT(data->castle_info) ) {
									data1->hashKey ^= zobrist::blackCastleRight;
								}
							}
						}
						else {

							uint64 source = (*m)&(data->is_my);
							uint64 dest = (*m)&(~(data->is_my));

							#ifndef EVAL_STATIC

							newDynamicValue = dynamicValue;
							if( (sourcePiece != whiteKingIndex) 
								&& (sourcePiece != blackKingIndex)
							/*	&& (sourcePiece != whitePawnIndex) 
								&& (sourcePiece != blackPawnIndex)*/ )
							{
								if( (!(source & CENTER_SQ)) && (dest & CENTER_SQ) ) {
									newDynamicValue += evaluation::CENTER_SQ_VALUE;
								}
								else {
									if( (source & CENTER_SQ) && (!(dest & CENTER_SQ)) ) {
										newDynamicValue -= evaluation::CENTER_SQ_VALUE;
									}
								}
							}

							#endif

							if( ((*m) & M_ENP) == M_ENP ) {
								#ifndef EVAL_STATIC

								newDynamicValue +=
									+ evaluation::PAWN_VALUE
									+ evaluation::PAWN_RANK6_VALUE
									- evaluation::PAWN_RANK5_VALUE;

								#endif
								int s, d;
								s = LS1B_INDEX((*m)&(~M_ENP)&(data->is_my));
								d = LS1B_INDEX((*m)&(~M_ENP)&(~(data->is_my)));
								data1->hashKey = zobrist::delta(data->hashKey,
											   LS1B_INDEX((*m)&(data->is_my)),
											   LS1B_INDEX((*m)&(~(data->is_my))),
											   sourcePiece,
											   destPiece,
											   data->next);
								if( data1->next == BLACK ) {
									data1->hashKey ^=
										zobrist::RandomPiece[
											zobrist::whitePawnIndex*64 + 8*(7-data->ep/8) + data->ep%8];
								}
								else {
									data1->hashKey ^=
										zobrist::RandomPiece[zobrist::blackPawnIndex*64 + data->ep];
								}
							}
							else {

								#ifndef EVAL_STATIC

								if( destPiece < noPieceIndex ) {
									if( dest & CENTER_SQ ) {
										newDynamicValue += evaluation::CENTER_SQ_VALUE;
									}
								}
								switch( destPiece ) {
									case whitePawnIndex:
									case blackPawnIndex:
										if( dest & ROW_2 ) {
											newDynamicValue +=
												+ evaluation::PAWN_VALUE
												+ evaluation::PAWN_RANK7_VALUE;
										}
										else {
											if( dest & ROW_3 ) {
												newDynamicValue +=
													+ evaluation::PAWN_VALUE
													+ evaluation::PAWN_RANK6_VALUE;
											}
											else {
												if( dest & ROW_4 ) {
													newDynamicValue +=
														+ evaluation::PAWN_VALUE
														+ evaluation::PAWN_RANK5_VALUE;
												}
												else {
													newDynamicValue +=
														+ evaluation::PAWN_VALUE;
												}
											}
										}
										break;
									case whiteBishopIndex:
									case blackBishopIndex:
										newDynamicValue +=
											+ evaluation::BISHOP_VALUE;
									/*	if( dest & CENTER_SQ ) {
											newDynamicValue += evaluation::CENTER_SQ_VALUE;
										}*/
										break;
									case whiteRookIndex:
									case blackRookIndex:
										newDynamicValue +=
											+ evaluation::ROOK_VALUE;
									/*	if( dest & CENTER_SQ ) {
											newDynamicValue += evaluation::CENTER_SQ_VALUE;
										}*/
										break;
									case whiteKnightIndex:
									case blackKnightIndex:
										newDynamicValue +=
											+ evaluation::KNIGHT_VALUE;
									/*	if( dest & CENTER_SQ ) {
											newDynamicValue += evaluation::CENTER_SQ_VALUE;
										}*/
										break;
									case whiteQueenIndex:
									case blackQueenIndex:
										newDynamicValue +=
											+ evaluation::QUEEN_VALUE;
									/*	if( dest & CENTER_SQ ) {
											newDynamicValue += evaluation::CENTER_SQ_VALUE;
										}*/
										break;
									
								}
								#endif

								if( data->next == WHITE ) {
									if( state_castleinfo::GET_MY_CASTLE_LEFT(data->castle_info)
										&& !state_castleinfo::GET_MY_CASTLE_LEFT(tCastleInfo) )
									{
										#ifndef EVAL_STATIC
										newDynamicValue += LOSE_RIGHT_TO_CASTLE_PENALTY_VALUE;
										#endif

										data1->hashKey ^= zobrist::whiteCastleLeft;
									}
									if( state_castleinfo::GET_MY_CASTLE_RIGHT(data->castle_info) 
										&& !state_castleinfo::GET_MY_CASTLE_RIGHT(tCastleInfo) )
									{
										#ifndef EVAL_STATIC
										newDynamicValue += LOSE_RIGHT_TO_CASTLE_PENALTY_VALUE;
										#endif
										data1->hashKey ^= zobrist::whiteCastleRight;
									}
									if( state_castleinfo::GET_OP_CASTLE_LEFT(data->castle_info)
										&& !state_castleinfo::GET_OP_CASTLE_LEFT(tCastleInfo) )
									{
										#ifndef EVAL_STATIC
										newDynamicValue -= LOSE_RIGHT_TO_CASTLE_PENALTY_VALUE;
										#endif

										data1->hashKey ^= zobrist::blackCastleLeft;
									}
									if( state_castleinfo::GET_OP_CASTLE_RIGHT(data->castle_info) 
										&& !state_castleinfo::GET_OP_CASTLE_RIGHT(tCastleInfo) )
									{
										#ifndef EVAL_STATIC
										newDynamicValue -= LOSE_RIGHT_TO_CASTLE_PENALTY_VALUE;
										#endif
										data1->hashKey ^= zobrist::blackCastleRight;
									}
								}
								else {
									if( state_castleinfo::GET_MY_CASTLE_LEFT(data->castle_info)
										&& !state_castleinfo::GET_MY_CASTLE_LEFT(tCastleInfo) )
									{
										#ifndef EVAL_STATIC
										newDynamicValue += LOSE_RIGHT_TO_CASTLE_PENALTY_VALUE;
										#endif
										data1->hashKey ^= zobrist::blackCastleLeft;
									}
									if( state_castleinfo::GET_MY_CASTLE_RIGHT(data->castle_info)
										&& !state_castleinfo::GET_MY_CASTLE_RIGHT(tCastleInfo) )
									{
										#ifndef EVAL_STATIC
										newDynamicValue += LOSE_RIGHT_TO_CASTLE_PENALTY_VALUE;
										#endif
										data1->hashKey ^= zobrist::blackCastleRight;
									}
									if( state_castleinfo::GET_OP_CASTLE_LEFT(data->castle_info)
										&& !state_castleinfo::GET_OP_CASTLE_LEFT(tCastleInfo) )
									{
										#ifndef EVAL_STATIC
										newDynamicValue -= LOSE_RIGHT_TO_CASTLE_PENALTY_VALUE;
										#endif
										data1->hashKey ^= zobrist::whiteCastleLeft;
									}
									if( state_castleinfo::GET_OP_CASTLE_RIGHT(data->castle_info)
										&& !state_castleinfo::GET_OP_CASTLE_RIGHT(tCastleInfo) )
									{
										#ifndef EVAL_STATIC
										newDynamicValue -= LOSE_RIGHT_TO_CASTLE_PENALTY_VALUE;
										#endif
										data1->hashKey ^= zobrist::whiteCastleRight;
									}
								}

								if( ((*m) & M_PROMOTE_ANY) == M_PROMOTE_ANY) {
									int s, d, pr_pi;
									s = LS1B_INDEX((*m)&OPPONENT_LINES&(data->is_my));
									d = LS1B_INDEX((*m)&OPPONENT_LINES&(~(data->is_my)));
									switch( (*m) & M_PROMOTE_ALL ) {
										case M_PROMOTE_Q:
											#ifndef EVAL_STATIC

											newDynamicValue += 
												- evaluation::PAWN_VALUE
												- evaluation::PAWN_RANK7_VALUE
												+ evaluation::QUEEN_VALUE;

											#endif
											pr_pi = zobrist::whiteQueenIndex;
											break;
										case M_PROMOTE_K:
											#ifndef EVAL_STATIC

											newDynamicValue +=
												- evaluation::PAWN_VALUE
												- evaluation::PAWN_RANK7_VALUE
												+ evaluation::KNIGHT_VALUE;

											#endif
											pr_pi = zobrist::whiteKnightIndex;
											break;
										case M_PROMOTE_R:
											#ifndef EVAL_STATIC

											newDynamicValue +=
												- evaluation::PAWN_VALUE
												- evaluation::PAWN_RANK7_VALUE
												+ evaluation::ROOK_VALUE;

											#endif
											pr_pi = zobrist::whiteRookIndex;
											break;
										case M_PROMOTE_B:
											#ifndef EVAL_STATIC

											newDynamicValue +=
												- evaluation::PAWN_VALUE
												- evaluation::PAWN_RANK7_VALUE
												+ evaluation::BISHOP_VALUE;

											#endif
											pr_pi = zobrist::whiteBishopIndex;
											break;
										default:
											assert( false );
											__assume( false );
											pr_pi = 0;
									}
									if( data->next == BLACK ) {
										++pr_pi;
									}
									data1->hashKey = 
										zobrist::delta(data1->hashKey,
													   s,
													   d,
													   sourcePiece,
													   destPiece,
													   data->next,
													   pr_pi);
								}
								else {
									#ifndef EVAL_STATIC
									switch( sourcePiece ) {
										case whitePawnIndex:
										case blackPawnIndex:
											if( dest & ROW_5 ) {
												newDynamicValue +=
													+ evaluation::PAWN_RANK5_VALUE;
											}
											else {
												if( dest & ROW_6 ) {
													newDynamicValue +=
														+ evaluation::PAWN_RANK6_VALUE
														- evaluation::PAWN_RANK5_VALUE;
												}
												else {
													if( dest & ROW_7 ) {
														newDynamicValue +=
															+ evaluation::PAWN_RANK7_VALUE
															- evaluation::PAWN_RANK6_VALUE;
													}
												}
											}
											break;
										case whiteKingIndex:
										case blackKingIndex:
											//newDynamicValue = dynamicValue + evaluation::QUEEN_VALUE;
											break;
										//default:
											//newDynamicValue = dynamicValue;

									}
									#endif
									data1->hashKey =
										zobrist::delta(data1->hashKey,
													   LS1B_INDEX(source),
													   LS1B_INDEX(dest),
													   sourcePiece,
													   destPiece,
													   data->next);
								}
							}
						}
					}
					assert( data->transpositionTable != 0 );
					#if ENGINE_LOG_LEVEL > 2
						uint64 logsource, logdest;
						if( data->next == WHITE ) {
							logsource = FLIPVERTICAL((*m)&data->is_my);
							logdest = FLIPVERTICAL((*m)&(~data->is_my));
						}
						else {
							logsource = ((*m)&data->is_my);
							logdest = ((*m)&(~data->is_my));
						}
						data->logMove.source = (BoardPosition)LS1B_INDEX(logsource);
						data->logMove.dest = (BoardPosition)LS1B_INDEX(logdest);
						data->logMove.einfo = NONE;
						Data* logD;
						logD = data;
						while( logD->all_p != 0 ) {
							--logD;
						}
						data->logger->pushWriteTimeFlag();
						data->logger->unsetWriteTime();
						(*(data->logger)) << "move treshold is: " << data->moveTreshold << " searching move: ";
						do {
							++logD;
							(*(data->logger)) << logD->logMove.print() << " - ";
						} while( logD != data );
						(*(data->logger)) << " alpha: " << data->alpha << " beta: " << data->beta << "\n";
						(*(data->logger)) << communicate::printBitBoard(data->all_p);
						(*(data->logger)) << communicate::printBitBoard(data->is_my);
						(*(data->logger)) << communicate::printBitBoard(data->is_op);
						data->logger->popWriteTimeFlag();
					#endif

					data1->beta = -data->alpha;

					#ifdef USE_PVS

					if( childrenSearched == 0 ) {
						data1->alpha = -data->beta;
					}
					else {
						data1->alpha = -data->alpha-1;
					}

					#else

					data1->alpha = -data->beta;

					#endif
					int val;
					assert( data->transpositionTable != 0 );

					bool newinLateMove;
					newinLateMove = false;

					#ifdef USE_LMR

					if( (ply > PLY_FULL_SEARCH_MIN + 2)
						&& (!(data->isPVNode || data1->isPVNode))
						&& (!checked)
						&& (
						  (
						   (m >= data->MoveList2 + checkingMovesIndex) 
						   && (childrenSearched > 2)
						  )
						  ||
					      (childrenSearched > 3)
						)
						#ifdef USE_VERIFYED_NULLMOVE
						&& nullVerify
						&& (!failHigh)
						#endif
						) 
					{
						--ply;
						newinLateMove = true;
						data1->alpha = -data->alpha-1;
					}

					#endif


					#ifdef USE_SIMPLE_QSEARCH
					if( ply > PLY_FULL_SEARCH_MIN ) {
					#endif
						val = -_search(
							data1,
							-newDynamicValue, 
							nodeCount, 
							nullVerify, 
							ply-1,
							newinLateMove|inLateMove);
					#ifdef USE_SIMPLE_QSEARCH
					}
					else {
	/*					if( nodeCount ==
						117876
						)
						{
							int kkokk;
							kkokk = 5;
						}
						uint64 kkkkkk = nodeCount;
*/						val = -_qsearch(
								data1->all_p,
								data1->is_my,
								data1->is_op,
								data1->is_pawn,
								data1->is_r,
								data1->is_b,
								data1->k_pos,
								-newDynamicValue,
								data1->alpha,
								data1->beta,
								nodeCount,
								0);
/*							int val22 = -_search(
								data1,
								-newDynamicValue, 
								nodeCount, 
								nullVerify, 
								ply-1,
								inLateMove);
							if( 
								(!
								  (
								    ( ( val <= data->alpha ) && (val22 <= data->alpha) )
									||
								    ( ( val >= data->beta ) && (val22 >= data->beta) )
								 )
								)
								&&
								(val != val22)
							/*	((val - val22) > 150)
								||
								((val - val22) < -150)
								
								) {
								int gdf;
								gdf = 1;
							}
*/						
					}
					#endif
					if( val == SEARCH_HALT_VALUE ) {
						return -SEARCH_HALT_VALUE;
					}
					if( val == SEARCH_ILLEGAL_MOVE_VALUE ) {
						++m;
						continue;
					}

					#ifdef USE_LMR

					if( newinLateMove ) {
						++ply;
					}

					#ifndef USE_PVS

					if( newinLateMove && (val > data->alpha) ) {
						data1->alpha = -data->beta;
						val = -_search(
							data1, 
							-newDynamicValue, 
							nodeCount, 
							nullVerify, 
							ply-1);
						if( val == SEARCH_HALT_VALUE ) {
							return -SEARCH_HALT_VALUE;
						}
					}

					#endif
					#endif

					#ifdef USE_PVS

					if( childrenSearched != 0 ) {
						#ifdef USE_LMR
						if( ((!newinLateMove) || (val < data->beta)) && (val > data->alpha) ) {
						#else
						if( (val > data->alpha) && (val < data->beta) ) {
						#endif
							data1->alpha = -data->beta;
							if( data->isPVNode ) {
								data1->isPVNode = true;
								data1->isAllNode = false;
								data1->isCutNode = false;
							}

							#ifdef USE_SIMPLE_QSEARCH
							if( ply > PLY_FULL_SEARCH_MIN ) {
							#endif
								val = -_search(
									data1,
									-newDynamicValue, 
									nodeCount, 
									nullVerify, 
									ply-1,
									inLateMove);
							#ifdef USE_SIMPLE_QSEARCH
							}
							else {
								val = -_qsearch(
										data1->all_p,
										data1->is_my,
										data1->is_op,
										data1->is_pawn,
										data1->is_r,
										data1->is_b,
										data1->k_pos,
										-newDynamicValue,
										data1->alpha,
										data1->beta,
										nodeCount,
										0);
	/*						int val22 = -_search(
								data1,
								-newDynamicValue, 
								nodeCount, 
								nullVerify, 
								ply-1,
								inLateMove);
							if( val != val22 ) {
								int gdf;
								gdf = 1;
							}
*/
							}
							#endif
							if( val == SEARCH_HALT_VALUE ) {
								return -SEARCH_HALT_VALUE;
							}
						}
					}

					#endif

				//	data1->isPVNode = false;

					assert( val < INFINITE_VALUE );
					assert( val > -INFINITE_VALUE );

					++childrenSearched;
					assert( data->transpositionTable != 0 );
					
					if( val > bestReturned ) {
						bestReturned = val;
						bestMove = *m;
						#ifdef PV_STORE
						data->bestLine[0] = *m;
						memcpy(&(data->bestLine[1]),
							   &(data1->bestLine[0]),
							   64*sizeof(uint64));
						#endif
					}
					if( val > data->alpha ) {
						isUpperBound = false;
						data->alpha = val;
						if( data->alpha >= data->beta ) {
					//		if( (*m) == data->killer2 ) {
					//			++CoUnTeR;
					//		}

							#ifdef USE_COUNTER_MOVES

							if( data->prevMove == 0 ) {
								data->nullCounterMove = (*m);
								if( (data+2)->nullCounterMove == 0 ) {
									(data+2)->nullCounterMove = (*m);
								}
								if( (data+4)->nullCounterMove == 0 ) {
									(data+4)->nullCounterMove = (*m);
								}
							}
							else {
								if( data->prevMove == CL_SQUARES ) {
									data->clCounterMove = (*m);
									if( (data+2)->clCounterMove == 0 ) {
										(data+2)->clCounterMove = (*m);
									}
									if( (data+4)->clCounterMove == 0 ) {
										(data+4)->clCounterMove = (*m);
									}
								}
								else {
									if( data->prevMove == CR_SQUARES ) {
										data->crCounterMove = (*m);
										if( (data+2)->crCounterMove == 0 ) {
											(data+2)->crCounterMove = (*m);
										}
										if( (data+4)->crCounterMove == 0 ) {
											(data+4)->crCounterMove = (*m);
										}
									}
									else {
										if( (data->prevMove&M_PROMOTE_ANY) == M_PROMOTE_ANY ) {
											data->prevMove = data->prevMove^M_PROMOTE_ANY;
										}
										if( (data->prevMove&M_ENP) == M_ENP ) {
											data->prevMove = data->prevMove^M_ENP;
										}
										int ps, pd;
										ps = LS1B_INDEX(((data-1)->is_my)&data->prevMove);
										pd = LS1B_INDEX((~((data-1)->is_my))&data->prevMove);
										data->counterMoveTable[ps][pd] = (*m);
										if( (data+2)->counterMoveTable[ps][pd] == 0 ) {
											(data+2)->counterMoveTable[ps][pd] = (*m);
										}
										if( (data+4)->counterMoveTable[ps][pd] == 0 ) {
											(data+4)->counterMoveTable[ps][pd] = (*m);
										}
									}
								}
							}

							#endif

							#if KILLER_MOVE_COUNT > 0

							if( (((*m)&M_PROMOTE_ANY) != M_PROMOTE_ANY)
								&& (((*m)&M_ENP) != M_ENP)
								&& (((*m) & data->is_op) == 0)
								)
							{

								#if KILLER_MOVE_COUNT == 1
								data->killer = (*m);
								#endif

								#if KILLER_MOVE_COUNT == 2
								if( data->killer != (*m) ) {
									data->killer1 = data->killer;
									data->killer = (*m);
								}
								#endif

								#if KILLER_MOVE_COUNT == 3
								if( (data->killer != (*m)) && (data->killer1 != (*m)) ) {
									data->killer2 = data->killer;
									data->killer1 = data->killer;
									data->killer = *m;
								}
								#endif

							}

							#endif

							if( (m+1) != 0 ) {
								isLowerBound = true;
							}
							goto SEARCH_RETURN;
						}
					}
					assert( data->transpositionTable != 0 );
					++m;

					if( data->isPVNode ) {
						data1->isPVNode = false;
						data1->isAllNode = false;
						data1->isCutNode = true;
					}
					else {
						if( data->isCutNode ) {
							if( m > data->MoveList+5 ) {
								data->isCutNode = false;
								data->isAllNode = true;
								data1->isCutNode = true;
								data1->isAllNode = false;
							}
							else {
								data1->isCutNode = false;
								data1->isAllNode = true;
							}
						}
					}

				} while( (*m) != 0 );

				SEARCH_RETURN:

				if( childrenSearched == 0 ) {
					if( allChildren ) {
						if( checked ) {
							// mate detection
							#ifdef USE_TP
							data->transpositionTable->insert(
								data->hashKey,
								0,
								0,
								-(evaluation::INFINITE_VALUE-1),
								false,
								false,
								31);
							#endif
							return -(evaluation::INFINITE_VALUE-100+(int)ply);
						}
						else {
							// stalemate detedtion
							#ifdef USE_TP
							data->transpositionTable->insert(
								data->hashKey,
								0,
								0,
								0,
								false,
								false,
								31);
							#endif
							return 0;
						}
					}
					else {
						#ifdef EVAL_STATIC

						if( !haveStaticValue ) {
							staticValue = evaluation::staticValue(data_BB_ARGUMENTS_CALL);
						}
						return staticValue;

						#else

						return dynamicValue;

						#endif
					}
				}

				#ifdef USE_VERIFYED_NULLMOVE
				if( failHigh && (bestReturned < data->beta) ) {
					//  zugzwang decetion
					data->alpha = originalAlpha;
					m = originalM;
					failHigh = false;
					nullVerify = true;
					ply = originalPly;
					underReSearch = true;
					goto RE_SEARCH;
				}
				#endif

				assert( data->transpositionTable != 0 );
				#ifdef USE_TP
				#ifdef USE_QTP
				if( (ply >= PLY_FULL_SEARCH_MIN - 16)
				#else
				if( (ply > PLY_FULL_SEARCH_MIN)
				#endif
					//&& (data->bestLine[0] != 0) )
					&& (bestMove != 0) )
				{
					Move move;
					//move.fromBitBoard(data->bestLine[0], data->is_my);
					move.fromBitBoard(bestMove, data->is_my);
					data->transpositionTable->insert(
						data->hashKey,
						move.source,
						move.dest,
						bestReturned,
						isUpperBound,
						isLowerBound,
						ply-(PLY_FULL_SEARCH_MIN-16),
						nullVerify);
					
					#ifdef _DEBUG					
					data->transpositionTable->lookup(data->hashKey, data->transponsEntry);
					//assert( ply-16 == data->transponsEntry.ply );
						uint64 m2;
						int val2;
						bool upper2, lower2;
						int ply2;
						data->transpositionTable->lookup(data->hashKey, data->transponsEntry);
						if( data->transponsEntry.searchValue == evaluation::SEARCH_ILLEGAL_MOVE_VALUE ) {
							++ply;
							--ply;
						}
						assert( data->transponsEntry.ply+(PLY_FULL_SEARCH_MIN-16) == ply );
						assert( data->transponsEntry.searchValue == bestReturned );
						assert( isUpperBound == data->transponsEntry.isUpperBound );
						assert( isLowerBound == data->transponsEntry.isLowerBound );
					//	m2 = (U64(1)<<data->transponsEntry.source) | (U64(1)<<data->transponsEntry.dest);
					//	assert( bestMove == m2 );
					#endif
				}
				#endif

				assert( data->transpositionTable != 0 );
				assert( bestReturned < evaluation::INFINITE_VALUE );
				assert( bestReturned > -evaluation::INFINITE_VALUE );

				if( allChildren || (bestReturned > dynamicValue) )  {
					return bestReturned;
				}
				else {
					return dynamicValue;
				}

			#	undef data_BB_ARGUMENTS_CALL
			} // Searcher::_search

		} // search
	} // engine
} // zochova
