
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#pragma once


// from http://www.velocityreviews.com/forums/t317511-binary-constant-macros.html


/* Binary constant generator macro
   By Tom Torfs - donated to the public domain
*/

/* All macro's evaluate to compile-time constants */

/* *** helper macros *** /

/* turn a numeric literal into a hex constant
   (avoids problems with leading zeroes)
   8-bit constants max value 0x11111111, always fits in unsigned long
*/
#define HEX__(n) 0x##n##LU

/* 8-bit conversion function */
#define B8__(x) ((x&0x0000000FLU)?1:0)      \
               +((x&0x000000F0LU)?2:0)      \
               +((x&0x00000F00LU)?4:0)      \
               +((x&0x0000F000LU)?8:0)      \
               +((x&0x000F0000LU)?16:0)     \
               +((x&0x00F00000LU)?32:0)     \
               +((x&0x0F000000LU)?64:0)     \
               +((x&0xF0000000LU)?128:0)

/* *** user macros *** /

/* for upto 8-bit binary constants */
#define B8(d) ((unsigned char)B8__(HEX__(d)))

/* for upto 16-bit binary constants, MSB first */
#define B16(dmsb,dlsb) (((unsigned short)B8(dmsb)<<8) \
                        + B8(dlsb))

/* for upto 32-bit binary constants, MSB first */
#define B32(dmsb,db2,db3,dlsb) (((unsigned long)B8(dmsb)<<24) \
                                  + ((unsigned long)B8(db2)<<16) \
                                  + ((unsigned long)B8(db3)<<8) \
                                  + B8(dlsb)) 

/* for upto 64-bit binary constants, MSB first */
#define B64(dmsb,db2,db3,db4,db5,db6,db7,dlsb) \
	(((uint64)B8(dmsb)<<56) \
    + ((uint64)B8(db2)<<48) \
    + ((uint64)B8(db3)<<40) \
    + ((uint64)B8(db4)<<32) \
    + ((uint64)B8(db5)<<24) \
    + ((uint64)B8(db6)<<16) \
    + ((uint64)B8(db7)<<8)  \
    + B8(dlsb)) 


/* 8-bit conversion function LS1B first */
#define B8__L(x) ((x&0x0000000FLU)?128:0)      \
               +((x&0x000000F0LU)?64:0)      \
               +((x&0x00000F00LU)?32:0)      \
               +((x&0x0000F000LU)?16:0)      \
               +((x&0x000F0000LU)?8:0)     \
               +((x&0x00F00000LU)?4:0)     \
               +((x&0x0F000000LU)?2:0)     \
               +((x&0xF0000000LU)?1:0)

/* for upto 8-bit binary constants LS1B first */
#define B8L(d) ((unsigned char)B8__L(HEX__(d)))

/* for upto 64-bit binary constants, LS1B first */
#define B64L(dlsb,db2,db3,db4,db5,db6,db7,dmsb) \
	(((uint64)B8L(dmsb)<<56) \
    + ((uint64)B8L(db7)<<48) \
    + ((uint64)B8L(db6)<<40) \
    + ((uint64)B8L(db5)<<32) \
    + ((uint64)B8L(db4)<<24) \
    + ((uint64)B8L(db3)<<16) \
    + ((uint64)B8L(db2)<<8)  \
    + B8L(dlsb))

