
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#pragma once

#include "State.h"

#define BB_ARGUMENTS \
uint64 all_p, \
uint64 is_my, \
uint64 is_op, \
uint64 is_pawn, \
uint64 is_r, \
uint64 is_b, \
uint64 k_pos

#define BB_DECLARE \
uint64 all_p; \
uint64 is_my; \
uint64 is_op; \
uint64 is_pawn; \
uint64 is_r; \
uint64 is_b; \
uint64 k_pos;

#define REFBB_ARGUMENTS \
uint64 &all_p, \
uint64 &is_my, \
uint64 &is_op, \
uint64 &is_pawn, \
uint64 &is_r, \
uint64 &is_b, \
uint64 &k_pos

#define BB_ARGUMENTS_CALL \
all_p, \
is_my, \
is_op, \
is_pawn, \
is_r, \
is_b, \
k_pos


namespace zochova
{
	//
	//	generating moves by bitboards, 
	//	initialize() must be called before using, it generates Move tables for king, knights
	namespace bitboardmoves
	{
		extern uint64 knight_moves[64];
		extern uint64 king_moves[64];

		void initialize();

		// list of my moves, without considering checked king, and castleing
		uint64 gen(BB_ARGUMENTS, uint64 *list, uint64 **listEnd = 0, bool plausibles = false);

		uint64 genCaptures(BB_ARGUMENTS, uint64 *list);

		//  generate enpassant if possible
		void genEnp(uint64 is_my, uint64 is_pawn, int ep, uint64 *list);

		//  verify if squares used by king area attacked
		//  and if not add castle move to move list
		void genCastleLeft(BB_ARGUMENTS, uint64 *move);
		void genCastleRight(BB_ARGUMENTS, uint64 *move);

		// checks if my king is attacked by opponent
		bool kingChecked(BB_ARGUMENTS, bool isFlipped = false);

		// checks if my king is attacked by opponent,
		// also returns possible attack squares
		bool kingCheckSquares(
			BB_ARGUMENTS,
			uint64 &att_p,
			uint64 &att_b,
			uint64 &att_r,
			uint64 &att_k,
			uint64 &att_disc);

		void makeMoveFlipped(REFBB_ARGUMENTS,
					  uint64 m,
					  int &sourcePieceIndex,
					  int &destPieceIndex,
					  state_castleinfo::CastleInfo &castleInfo,
					  Side turn,
					  int &ep);

		int mobility(BB_ARGUMENTS);

	}
};
