
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "communicate.h"
#include "comm_interface_list.h"
#include "log.h"

#include <iostream>

using namespace std;

namespace zochova
{
	namespace communicate
	{

		void stringSplit(string str, string delim, vector<string>& results) {
			size_t cutAt;
			while( (cutAt = str.find_first_of(delim)) != str.npos ) {
				if(cutAt > 0) {
					results.push_back(str.substr(0, cutAt));
				}
				str = str.substr(cutAt+1);
			}
			if(str.length() > 0) {
				results.push_back(str);
			}
		}

		void ICommunication::setLogWriter(log::IWriter* logw) {
			logger = logw;
		}

		ICommunication* ICommunication::Factory() {
			string str;
			ICommunication* commi;
			cout << ZOCHOVA_NAME << " " << ZOCHOVA_VERSION << " type help for help" << endl;
			cin >> str;
			if( str == "xboard" ) {
				commi = new CommXboard();
			}
			else {
				if( str == "uci" ) {
					/*
						commi = new comm_uci();
						UCI not supported currently
					*/
					commi = new CommConsole(str);
				}
				else {
					commi = new CommConsole(str);
				}
			}
			return commi;
		}

		string printBitBoard(uint64 bb) {
			string s;
			uint64 bit;
			int i, j;
			bit = 1;
			j = 8;
			while( j-- ) {
				s.push_back('\n');
				i = 8;
				while( i-- ) {
					s.push_back((bb&bit)?'1':'0');
					bit <<= 1;
				}
			}
			s.push_back('\n');
			return s;
		}

		char printSquare(Square p) {
			switch(p) {
				case WHITE_PAWN: return 'P';
				case WHITE_QUEEN: return 'Q';
				case WHITE_KING: return 'K';
				case WHITE_BISHOP: return 'B';
				case WHITE_KNIGHT: return 'N';
				case WHITE_ROOK: return 'R';
				case BLACK_PAWN: return 'p';
				case BLACK_QUEEN: return 'q';
				case BLACK_KING: return 'k';
				case BLACK_BISHOP: return 'b';
				case BLACK_KNIGHT: return 'n';
				case BLACK_ROOK: return 'r';
				case EMPTY: return ' ';
				default: return '?';
			}
		}

		bool readMove(Move &m, string str) {
			if( (str.length() < 4) || (str.length() > 5) ) {
				return false;
			}
			char move_str[6];
			strcpy_s(move_str, 6, str.c_str());
			if( (move_str[0] < 'I') && (move_str[0] >= 'A') ) {
				move_str[0] -= ('A'-'a');
			}
			if( (move_str[2] < 'I') && (move_str[2] >= 'A') ) {
				move_str[2] -= ('A'-'a');
			}
			if( !((move_str[0] < 'i') && (move_str[0] >= 'a') &&
				 (move_str[1] < '9') && (move_str[1] >= '1') &&
				 (move_str[2] < 'i') && (move_str[2] >= 'a') &&
				 (move_str[3] < '9') && (move_str[3] >= '1')
			)) {
				return false;
			}
			m.source = (BoardPosition)(move_str[0]-'a'+(7-(move_str[1]-'1'))*8);
			m.dest = (BoardPosition)(move_str[2]-'a'+(7-(move_str[3]-'1'))*8);
			switch( move_str[4] ) {
				case 'q':
				case 'Q':
					m.einfo = PROMOTE;
					m.promote = QUEEN;
					break;
				case 'n':
				case 'N':
					m.einfo = PROMOTE;
					m.promote = KNIGHT;
					break;
				case 'r':
				case 'R':
					m.einfo = PROMOTE;
					m.promote = ROOK;
					break;
				case 'b':
				case 'B':
					m.einfo = PROMOTE;
					m.promote = BISHOP;
					break;
				default:
					m.einfo = NONE;
					break;
			}
			return true;
		} // readMove

		void ICommunication::doInputThread() {
			string command;
			//for(;;) {
			while( true) {
				getline(cin, command);
				if( logger != NULL ) {
					(*logger) << logInputPrefix << command << "\n";
				}
				if( command.length() > 0 ) { 
					processed = false;
					if( !newInput(command) ) {
						return;
					}
				}
			}
		}

		void ICommunication::run() {
			processed = true;
			doInputThread();
		}

	} // communicate
} // zochova
