
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "log.h"
#include "include_boost_datetime.h"


namespace zochova
{
	namespace log
	{
		IWriter::IWriter() {
			isOpen = false;
			WriteTimeFlag = false;
		}

		void IWriter::pushWriteTimeFlag() {
			writeTimeFlagStack.push(WriteTimeFlag);
		}

		void IWriter::popWriteTimeFlag() {
			WriteTimeFlag = writeTimeFlagStack.top();
			writeTimeFlagStack.pop();
		}

		void IWriter::setWriteTime() {
			WriteTimeFlag = true;
		}

		void IWriter::unsetWriteTime() {
			WriteTimeFlag = false;
		}

		IWriter& RealWriter::write(std::string str) {
			if( afterNewLine && WriteTimeFlag ) {
				fileStream << "[" << boost::posix_time::second_clock::local_time() << "] ";
			}
			fileStream << str;
			fileStream.flush();
			if( str.find("\n") == std::string::npos ) {
				afterNewLine = false;
			}
			else {
				fileStream.flush();
				afterNewLine = true;
			}
			return *this;
		}

		RealWriter::RealWriter() {
			isOpen = false;
			WriteTimeFlag = true;
			afterNewLine = true;
		}

		RealWriter::~RealWriter() {
			fileStream.close();
		}

		RealWriter::RealWriter(std::string argFileName, bool append) {
			fileName = argFileName;
			if( append ) {
				fileStream.open(fileName.c_str(), std::ios::app);
			}
			else {
				fileStream.open(fileName.c_str());
			}
			fileStream
				<< "log by " << ZOCHOVA_NAME << " " << ZOCHOVA_VERSION
				<< " build " << __DATE__ << " compiler: " << ZOCHOVA_COMPILER_NAME << " " << ZOCHOVA_COMPILER_VERSION
				<< " target platform: " << ZOCHOVA_PLATFORM_STRING << "\n";
			isOpen = true;
			WriteTimeFlag = true;
		}

	} // log
} // zochova
