
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

//
//  platform and compiler specific things,
//  and some magic from http://chessprogramming.wikispaces.com/
//

#pragma once


#ifdef _MSC_VER
#	define _CRTDBG_MAP_ALLOC
#	include <stdlib.h>
#	pragma warning(push)
#	pragma warning( disable : 4005 )
#	include <crtdbg.h>
#	pragma warning(pop)
#	include <CodeAnalysis\Warnings.h>
#	include <Codeanalysis\sourceannotations.h>
#	pragma warning(push, 4)
#	include <intrin.h>
	typedef unsigned __int64 uint64;
#	define U64(u) (u##ui64)
#	define ZOCHOVA_COMPILER_NAME "MSVC"
#	define ZOCHOVA_COMPILER_VERSION _MSC_VER
#elif defined(__GNUC__)
#	define _CrtDumpMemoryLeaks() ((void)0);
#	include "gcc_intrin.h"
	typedef unsigned long long int uint64;
#	define U64(u) (u##ULL)
#	define ZOCHOVA_COMPILER_NAME "GCC"
#	define ZOCHOVA_COMPILER_VERSION __GNUC__
#else
	error unsupported / untested compiler
#endif

#ifndef __assume
#define __assume(thing) ((void) (thing))
#endif

#ifndef __analysis_assume
#define __analysis_assume(thing) ((void) (thing))
#endif

#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__)
#	if defined(WIN64) || defined(_WIN64) || defined(__WIN64__)
#		define ZOCHOVA_PLATFORM_64
#		define ZOCHOVA_PLATFORM_STRING "Windows 64bit"
#	else
#		define ZOCHOVA_PLATFORM_32
#		define ZOCHOVA_PLATFORM_STRING "Windows 32bit"
#	endif
#	define ZOCHOVA_PLATFOTM_WINDOWS
#elif defined(__gnu_linux__) || defined(linux) || defined(__linux) || defined(__linux__)
#	define ZOCHOVA_PLATFORM_32
#	define ZOCHOVA_PLATFOTM_LINUX
#	define ZOCHOVA_PLATFORM_STRING "Linux 32bit"
#elif defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
#	define ZOCHOVA_PLATFORM_32
#	define ZOCHOVA_PLATFOTM_BSD
#	define ZOCHOVA_PLATFORM_STRING "BSD 32bit"
#else
	error unsupported / untested platform
#endif

#include "flags.h"

#include "assert.h"
#include "binary.h"


namespace zochova
{


	#ifdef ZOCHOVA_PLATFORM_64

	#define GET_LS1B_INDEX(index, mask) _BitScanForward64(index, mask)

	inline static unsigned long LS1B_INDEX(uint64 mask) {
		unsigned long index;
		_BitScanForward64(&index, mask);
		return index;
	}

	#define GET_MS1B_INDEX(index, mask) _BitScanReverse64(index, mask)

	inline static unsigned long MS1B_INDEX(uint64 mask) {
		unsigned long	 index;
		_BitScanReverse64(&index, mask);
		return index;
	}

	#define ROL(VAL64, X) _rotl64((VAL64), (X))
	#define ROR(VAL64, X) _rotr64((VAL64), (X))

	#define FLIPVERTICAL(VAL64) _byteswap_uint64(VAL64)


	#else // ifdef ZOCHOVA_PLATFORM_64

	inline static unsigned char GET_LS1B_INDEX(unsigned long *index, uint64 mask) {
		if( mask & U64(0x00000000FFFFFFFF) ) {
			return _BitScanForward(index, (unsigned long)mask);
		}
		if( !_BitScanForward(index, mask>>32) ) {
			return 0;
		}
		(*index) += 32;
		return 1;
	}

	inline static unsigned long LS1B_INDEX(uint64 mask) {
		unsigned long index;
		if( _BitScanForward(&index, (unsigned long)mask) ) {
			return index;
		}
		_BitScanForward(&index, mask>>32);
		return index+32;
	}


	#define ROL(VAL64, X) (((VAL64) << X) | ((VAL64) >> (64-X)))
	#define ROR(VAL64, X) (((VAL64) >> X) | ((VAL64) << (64-X)))

	inline static uint64 FLIPVERTICAL(uint64 x) {
		#define k1 U64(0x00FF00FF00FF00FF)
		#define k2 U64(0x0000FFFF0000FFFF)
		#define k4 U64(0x00000000FFFFFFFF)
		x = ((x >>  8) & k1) | ((x & k1) <<  8);
		x = ((x >> 16) & k2) | ((x & k2) << 16);
		x = ((x >> 32) & k4) | ((x & k4) << 32);
		return x;
		#undef k1
		#undef k2
		#undef k4
	}

	#endif // ifdef ZOCHOVA_PLATFORM_64


	#ifdef __popcnt

	#if ZOCHOVA_PLATFORM_64
	#	define POPCOUNT(mask) __popcnt64(mask)
	#else
	#	define POPCOUNT(mask) (__popcnt((unsigned int)mask)+__popcnt((unsigned int)mask>>32))
	#endif

	#else 

	inline static int POPCOUNT(uint64 x) {
		x =  x       - ((x >> 1)  & U64(0x5555555555555555));
		x = (x & U64(0x3333333333333333)) + ((x >> 2)  & U64(0x3333333333333333));
		x = (x       +  (x >> 4)) & U64(0x0f0f0f0f0f0f0f0f) ;
		x = (x * U64(0x0101010101010101)) >> 56;
		return (int) x;
	}

	#endif

	inline static uint64 LS1B(uint64 x) {
		return ((x)&(U64(0)-(x)));
	}

	inline static bool POPCOUNT_MAX1(uint64 x) {
		return (x & (x-1)) == 0;
	}

	void delay(int csecs);

	class Timer
	{
	public:

		Timer();
		~Timer();

		int query();

	private:

		class TimerData;

		TimerData* startTime;

	};

} // zochova
