
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "search_flags.h"
#include "Searcher.h"
#include "evaluation.h"

namespace zochova
{
	namespace engine
	{
		namespace search
		{

			void Searcher::moveOrder(
				BB_ARGUMENTS, 
				uint64 * list,
				uint64 transposMove,
				uint64 killer,
				uint64 killer1,
				uint64 killer2)
			{
				(*(list+0)) = goto_next_move;
				(*(list+1)) = goto_next_move;
				(*(list+2)) = goto_next_move;
				(*(list+3)) = goto_next_move;
				(*(list+4)) = goto_next_move;

				uint64 * pBestMove;
				pBestMove = 0;
				int bestMoveValue;
				bestMoveValue = 0;

				uint64 * m;
				m = list + 5;
				do {
					if( (*m) == transposMove ) {
						(*(list+0)) = transposMove;
						(*m) = goto_next_move;
						++m;
						continue;
					}
					if( (*m) == killer ) {
						(*(list+1)) = killer;
						(*m) = goto_next_move;
						++m;
						continue;
					}
					if( (*m) == killer1 ) {
						(*(list+3)) = killer1;
						(*m) = goto_next_move;
						++m;
						continue;
					}
					if( (*m) == killer2 ) {
						(*(list+4)) = killer2;
						(*m) = goto_next_move;
						++m;
						continue;
					}
					if( ((*m) & M_PROMOTE_ANY) == M_PROMOTE_ANY ) {
						pBestMove = m;
						break;
					}
					if( ((*m) & M_ENP) == M_ENP ) {
						if( bestMoveValue < 2 ) {
							pBestMove = m;
							bestMoveValue = 2;
						}
					}
					else {
						if( ((*m) == CL_SQUARES) || ((*m) == CR_SQUARES) ) {
							if( bestMoveValue == 0 ) {
								pBestMove = m;
								bestMoveValue = 1;
							}
						}
						else {
							if( !((*m) & (is_op)) ) {
								++m;
								continue;
							}
							if( (*m) & is_op & is_r & is_b ) {
								if( (*m) & is_my & is_r & is_b ) {
									if( bestMoveValue < 8 ) {
										pBestMove = m;
										bestMoveValue = 8;
									}
								}
								else {
									pBestMove = m;
									bestMoveValue = 9;
								}
							}
							else {
								if( (*m) & is_op & is_pawn ) {
									if( bestMoveValue < 2 ) {
										pBestMove = m;
										bestMoveValue = 2;
									}
								}
								else {
									if( (*m) & is_my & is_pawn ) {
										if( bestMoveValue < 4 ) {
											pBestMove = m;
											bestMoveValue = 4;
										}
									}
									else {
										if( bestMoveValue < 3 ) {
											pBestMove = m;
											bestMoveValue = 3;
										}
									}
								}
							}
						}
					}
					++m;
				} while( (*m) != 0 );
				if( pBestMove == 0 ) {
					return;
				}
				(*(list+2)) = (*pBestMove);
				(*pBestMove) = goto_next_move;
				/*
				if( bestMoveValue == 1 ) {
					return;
				}
				uint64 * cap;
				cap = list + 5;
				m = list + 5;
				while( (*m) != 0 ) {
					if( ((*m) != goto_next_move)
						&&
						( (((*m) & M_PROMOTE_ANY) == M_PROMOTE_ANY)
						  || (((*m) & M_ENP) == M_ENP)
						  || ((*m) & is_op) ) )
					{
						uint64 t;
						t = (*m);
						(*m) = (*cap);
						(*cap) = t;
					}
					++m;
				}
*/
			}

			void Searcher::nextMove(
				BB_ARGUMENTS, 
				uint64 * list)
			{
				uint64 * pBestMove;
				pBestMove = 0;
				int bestMoveValue;
				bestMoveValue = 0;
				uint64 * m;
				m = list;
				do {
					if( (*m) == goto_next_move ) {
						++m;
						continue;
					}
					if( ((*m) & M_PROMOTE_ANY) == M_PROMOTE_ANY ) {
						pBestMove = m;
						break;
					}
					if( ((*m) & M_ENP) == M_ENP ) {
						if( bestMoveValue < 5 ) {
							pBestMove = m;
							bestMoveValue = 5;
						}
					}
					else {
						if( ((*m) == CL_SQUARES) || ((*m) == CR_SQUARES) ) {
							if( bestMoveValue == 0 ) {
								pBestMove = m;
								bestMoveValue = 1;
							}
						}
						else {
							if( !((*m) & (is_op)) ) {
								if( (*m) & is_my & is_pawn ) {
									if( bestMoveValue < 2 ) {
										pBestMove = m;
										bestMoveValue = 2;
									}
								}
								++m;
								continue;
							}
							if( (*m) & is_op & is_r & is_b ) {
								if( (*m) & is_my & is_r & is_b ) {
									if( bestMoveValue < 8 ) {
										pBestMove = m;
										bestMoveValue = 8;
									}
								}
								else {
									pBestMove = m;
									bestMoveValue = 9;
								}
							}
							else {
								if( (*m) & is_op & is_pawn ) {
									if( bestMoveValue < 5 ) {
										pBestMove = m;
										bestMoveValue = 5;
									}
								}
								else {
									if( (*m) & is_my & is_pawn ) {
										if( bestMoveValue < 7 ) {
											pBestMove = m;
											bestMoveValue = 7;
										}
									}
									else {
										if( bestMoveValue < 6 ) {
											pBestMove = m;
											bestMoveValue = 6;
										}
									}
								}
							}
						}
					}
					++m;
				} while( (*m) != 0 );
				if( (pBestMove == 0) || (pBestMove == list) ) {
					return;
				}
				uint64 t;
				t = (*list);
				(*list) = (*pBestMove);
				(*pBestMove) = t;
			}
/*
			bool Searcher::moveOrder(
				BB_ARGUMENTS,
				uint64* list,
				uint64* list2,
				uint64 transposMove,
				uint64 killer,
				uint64 killer1,
				uint64 killer2,
				uint64 is_op_major,
				uint64 is_op_pawn,
				uint64 att_p,
				uint64 att_b,
				uint64 att_r,
				uint64 att_k,
				uint64 att_disc,
				bool needLateMoves,
				bool needCheckingMoves)
			{
				static const int majorCaptureIndex = 4;
				static const int pawnCaptureIndex = 44;

#ifdef _DEBUG
 int x;
 x = 0;
 uint64 * xx = list;
 while( *xx ) {
	 ++xx;
	++x;
 }
 CoUnTeR2 = 0;
#endif

				(*(list2+0)) = goto_next_move;
				(*(list2+1)) = goto_next_move;
				(*(list2+2)) = goto_next_move;
				(*(list2+3)) = goto_next_move;

				uint64 * majorCap;
				uint64 * pawnCap;
				uint64 * checking;
				uint64 * others;

				majorCap = list2+majorCaptureIndex;
				pawnCap = list2+pawnCaptureIndex;
				checking = list2+checkingMovesIndex;
				others = list2+lateMovesIndex;

				uint64 * m;
				m = list;

				bool r;
				r = false;

				while( (*m) != 0 ) {
					++CoUnTeR2;
					if( (*m) == transposMove ) {
						(*(list2+0)) = transposMove;
						++m;
						r = true;
						continue;
					}
					if( (*m) == killer ) {
						(*(list2+1)) = killer;
						++m;
						r = true;
						continue;
					}
					if( (*m) == killer1 ) {
						(*(list2+2)) = killer1;
						++m;
						r = true;
						continue;
					}
				/*	if( (*m) == killer2 ) {
						(*(list2+3)) = killer2;
						++m;
						r = true;
						continue;
					}*/
			
		/*	
					if( (*m) == CL_SQUARES ) {
						if( SQ_D1 & att_r ) {
							if( needCheckingMoves ) {
								(*checking) = (*m);
								++checking;
								r = true;

							//	++CoUnTeR;

							}
							++m;
							continue;
						}
						if( needLateMoves ) {
							(*others) = (*m);
							++others;
							r = true;
						}
						++m;
						continue;
					}
					if( (*m) == CR_SQUARES ) {
						if( SQ_F1 & att_r ) {
							if( needCheckingMoves ) {

				//				++CoUnTeR;

								(*checking) = (*m);
								++checking;
								r = true;
							}
							++m;
							continue;
						}
						if( needLateMoves ) {
							(*others) = (*m);
							++others;
							r = true;
						}
						++m;
						continue;
					}
					if( ((*m) & M_ENP) != M_ENP ) {
						if( (((*m) & M_PROMOTE_ANY) == M_PROMOTE_ANY )
							||
							((*m) & is_op_major) )
						{
							r = true;
							(*majorCap) = (*m);
							++majorCap;
							++m;
							continue;
						}
					}
					if( (((*m) & M_ENP) == M_ENP )
						||
						((*m) & is_op_pawn) )
					{
						r = true;
						(*pawnCap) = (*m);
						++pawnCap;
						++m;
						continue;
					}
					if( ( (*m) & is_my & att_disc )
						||
						( ((*m) & is_r) && ((*m) & (~is_my) & att_r) ) 
						||
						( ((*m) & is_b) && ((*m) & (~is_my) & att_b) ) 
						||
						( ((*m) & is_pawn) && ((*m) & (~is_my) & att_p) ) 
						||
						( ((*m) & (~(is_pawn|is_b|is_r|is_pawn|k_pos))) && ((*m) & (~is_my) & att_k) ) )
					{
						if( needCheckingMoves ) {

				//			++CoUnTeR;

							r = true;
							(*checking) = (*m);
							++checking;
						}
						++m;
						continue;
					}
					if( needLateMoves ) {
						r = true;
						(*others) = (*m);
						++others;
					}
					++m;
				}
				if( !r ) {
					return false;
				}
				assert( majorCap < list2+pawnCaptureIndex );

	#ifdef _DEBUG
				int faszom = (list2+checkingMovesIndex)-pawnCap;
				int faszom2 = pawnCap-(list2+checkingMovesIndex);
		#endif
				assert( pawnCap < list2+checkingMovesIndex );


				assert( checking < list2+lateMovesIndex );
				while( majorCap < list2+pawnCaptureIndex ) {
					(*majorCap) = goto_next_move;
					++majorCap;
				}
				if( needCheckingMoves ) {
					while( pawnCap < list2+checkingMovesIndex ) {
						(*pawnCap) = goto_next_move;
						++pawnCap;
					}
					if( needLateMoves ) {
						(*others) = 0;
						while( checking < list2+lateMovesIndex ) {
							(*checking) = goto_next_move;
							++checking;
						}
					}
					else {
						(*checking) = 0;
					}
				}
				else {
					(*pawnCap) = 0;
				}
#ifdef _DEBUG
				if( needLateMoves ) {
					int z;
					 z = 0;
					 uint64 * zz = list2;
					 while( *zz ) {
						 if( (*zz) != goto_next_move ) {
							++z;
						 }
						 ++zz;
					 }
					 assert( z == x );
					 assert( z == CoUnTeR2 );
				}
#endif
				return true;
			}

/*

			void Searcher::sort_moves(uint64 *moves,
									  int *values,
									  bool *threats,
									  int length)
			{
				int t, i;
				uint64 tm;
				bool tb;
				while( length > 1 ) {
					i = 0;
					while( i < length-1 ) {
						if( values[i] < values[i+1] ) {
							t = values[i+1];
							tm = moves[i+1];
							tb = threats[i+1];
							values[i+1] = values[i];
							moves[i+1] = moves[i];
							threats[i+1] = threats[i];
							values[i] = t;
							moves[i] = tm;
							threats[i] = tb;
						}
						i++;
					}
					length--;
				}
			} // Searcher::sort_moves

			uint64* Searcher::moveOrder(uint64* list,
							BB_ARGUMENTS,
					//		uint64* killerMoves,
							int moveTreshold,
							uint64 transposMove,
							int* moveValues,
							bool * threats)
			{
				using namespace evaluation;
				#ifdef _DEBUG
				static int sum_kill = 0;
				static int count_kill = 0;
				#endif
				uint64 *m;
				uint64 s, d;
				//int moveValues[128];
				int *mv0;
				int *mv;
				int val;
				bool * thr;
			//	uint64 *killer;

				(void) all_p;
				//uint64 kills;
				//kills = 0;
				//killer = killerMoves;
				//while( (*killer) != 0 ) {
				//	kills |= *killer&is_my;
				//	++killer;
				//}
				m = list;
				mv0 = mv = &(moveValues[0]);
				thr = threats;
				while( (*m) != 0 ) {
					uint64 move = *m;
					(*thr) = false;
					if( (move) == transposMove ) {
						val = INFINITE_MOVE_VALUE;
						goto LOOP_END;
					}
					if( (move) == CR_SQUARES ) {
						val = CASTLE_RIGHT_VALUE;
						goto LOOP_END;
					}
					if( (move) == CL_SQUARES ) {
						val = CASTLE_LEFT_VALUE;
						goto LOOP_END;
					}
					if( ((move) & M_PROMOTE_ANY) == M_PROMOTE_ANY ) {
						val = PROMOTE_VALUE;
						goto LOOP_END;
					}
					val = 0;
					s = (move) & is_my;
					d = (move) - s;
					//if( s & kills ) {
					//	killer = killerMoves;
					//	#ifdef _DEBUG
					//	++count_kill;
					//	#endif
					//	while( (*killer) != 0 ) {
					//		#ifdef _DEBUG
					//		++sum_kill;
					//		#endif
					//		if( (move) == (*killer) ) {
					//			val = KILLER_VALUE;
					//			break;
					//		}
					//		++killer;
					//	}
					//}
					if( d & is_op ) {
						if( d & is_pawn ) {
							val += CAPTURE_PAWN_VALUE;
						}
						else {
							val += CAPTURE_NONPAWN_VALUE;
						}
					}
					if( s & k_pos ) {
						goto LOOP_END;
					}
					if( (d & CENTER_SQ) && !(s & CENTER_SQ) ) {
						val += TO_CENTER_VALUE;
					}
					if( s & is_r ) {
						if( (~s) & is_my & is_r & (EDGE_LEFT << (LS1B_INDEX(d)%8)) ) {
							val += ROOK_TO_ROOK_VALUE;
						}
					}
					else {
						if( s & is_pawn ) {
							if( (!(d&EDGE_RIGHT)) && ((d<<9)&is_pawn&is_my) ) {
								val += PAWN_TO_PAWN_VALUE;
							}
							if( (!(d&EDGE_LEFT)) && ((d<<7)&is_pawn&is_my) ) {
								val += PAWN_TO_PAWN_VALUE;
							}
							if( (!(d&(EDGE_LEFT|EDGE_RIGHT))) && ((d<<7)&((~is_pawn)&is_op)) && ((d<<9)&((~is_pawn)&is_op)) ) {
								(*thr) = true;
								val += FORK_VALUE;
							}
						}
						else {
							if( !(s&is_b) ) {
								if( POPCOUNT(bitboardmoves::knight_moves[LS1B_INDEX(d)] & is_op & (~is_pawn)) > 1 ) {
									(*thr) = true;
									val += FORK_VALUE;
								}
							}
						}
					}
					if( (d & OPPONENT_LINES) && !(s & OPPONENT_LINES) ) {
						if( s & is_pawn ) {
							val += PAWN_TO_OPPONENT_LINES_VALUE;
						}
						else {
							val += TO_OPPONENT_LINES_VALUE;
						}
					}
					LOOP_END:
					if( val < moveTreshold ) {
						if( list != m ) {
							(*m) = (*list);
							(*mv) = (*mv0);
							(*thr) = (*threats);
						}
						//(*list) = 0;
						++list;
						++mv0;
						++threats;
					}
					else {
						(*mv) = val;
					}
					++m;
					++mv;
					++thr;
				}
				sort_moves(list, mv0, threats, (int)(m-list));
				return list;
			} // Searcher::moveOrder
*/
		} // search
	} // engine
} // zochova
