
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#pragma once

#include "State.h"

#include "include_boost_bind_signals.h"
#include "include_boost_thread.h"

#include <vector>


namespace zochova
{

	class ChessGame
	{

	public:
		ChessGame();

		void reset();

		// make a Move, stoped player's clock, start opponent's clock
		void makeMove(const Move &m);
		

		// isStoped the clock, undo the latest Move
		bool undo();

		// if undo was done, get the next state in state history
		bool redo();

		// setup the clock
		void setTimeLimit(int movs,		// moves per time
						  int csecs,	// centiseconds for all the moves
						  int csecs_inc	// time to increment players clock on each Move
		);

		void setWhiteClock(int csecs);
		void setBlackClock(int csecs);

		void startTimer();

		void pauseTimer();

		void stopTimer();

		const State getCurrentState() const;

		void getCurrentState(State* state);

		void getClock(int* moves, int* csecs, int* opcsecs);

		boost::signal< void() > eventBlacksTurn;
		boost::signal< void() > eventWhitesTurn;
		boost::signal< void() > eventOfferDraw;
		boost::signal< void() > eventBlackResigns;
		boost::signal< void() > eventWhiteResigns;
		boost::signal< void(const Move &m) > eventNewMove;
		boost::signal< void() > eventIllegalMove;
		boost::signal< void() > eventAcceptedMove;
	private:
		
		//	All times are in centiseconds
		struct {
			// the time for moves for a player at the beginning of the game
			int gameTime;

			//  number of moves to make before time expires
			int gameMoves;

			// the number of centiseconds to incrase players time on each Move
			int gameTimeIncrementation;

			//  if true, time gameTime is added to clock after gameMoves
			bool gameTimeIncrementationGameMoves;

			// clock for sides
			int timeForWhite;
			int timeForBlack;

			// number of moves till time expires
			int movesForWhite;				
			int movesForBlack;

			// the clock of the player to Move when the turn began
			int thinkingStarted;

			// the player to Move
			Side thinkingSide;

			bool isPaused;
			bool isStoped;
		} timeInfo;

		class HistoryEntry {
		public:
			HistoryEntry() {
				
			};

			HistoryEntry(State &argState) {
				state = argState;
				move = nullMove;
			};

			HistoryEntry(State &argState, Move &argMove) {
				state = argState;
				move = argMove;
			};

			State state;
			Move move;
		};

		// each state and Move of the game
		std::vector< HistoryEntry > history;

		std::vector< HistoryEntry >::iterator currentState;

		boost::thread *clockThread;

		// the code that decreases the the clock of
		// the player to Move in every centisecond
		void doClockThread();
		
	}; // ChessGame

} // zochova
