
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "comm_interface_list.h"

#include <iostream>

#include "log.h"

using namespace std;
using boost::lexical_cast;

namespace zochova
{
	namespace communicate
	{

		CommConsole::CommConsole() {
			isShowBoard = true;
			isShowThinking = false;
			logInputPrefix = "from user: ";
			logger = NULL;
		}

		CommConsole::CommConsole(string first_command) {
			firstcommand = first_command;
			CommConsole();
		}

		void CommConsole::run() {
			if( newInput(firstcommand) ) {
				this->ICommunication::run();
			}
		}

		void CommConsole::sendOutput(string str) {
			if( logger != NULL ) {
				(*logger) << str << "\n";
			}
			cout << str << endl;
		}

		bool CommConsole::newInput(string command) {
			Move m;
			if( command.substr(0, 3) == "new" ) {
				eventNewGame();
			}
			else if( command.substr(0, 6) == "resign" ) {
				eventResign();
			}
			else if( command.substr(0, 4) == "quit" ) {
				eventQuit();
				return false;
			}
			else if( command.substr(0, 8) == "board on" ) {
				isShowBoard = true;
			}
			else if( command.substr(0, 9) == "board off" ) {
				isShowBoard = false;
			}
			else if( command.substr(0, 12) == "engine black" ) {
				eventSetEngineBlack();
			}
			else if( command.substr(0, 12) == "engine white" ) {
				eventSetEngineWhite();
			}
			else if( command.substr(0, 4) == "help" ) {
				sendOutput(
"Commands:\n\
new   - start new game\n\
engine white    - set engine to play white\n\
engine black    - set engine to play black\n\
undo   - undo previous Move\n\
redo   - redo next Move\n\
resign   - resign current game\n\
offer draw   - offer draw\n\
accept draw   - accept engine's draw offer\n\
quit   - quit zochova\n\
board on   - turn on echoing the current board after each Move\n\
board off   - turn off echoing the current board after each Move\n\
level moves time timeinc - set time control\
");
			}
			else if( command.substr(0, 4) == "ping" ) {
				sendOutput(command.replace(1, 1, "o"));		
			}
			else if( command.substr(0, 5) == "level" ) {
				vector<string> parts;
				stringSplit(command, " ", parts);
				if( parts.size() == 4 ) {
					try {
						eventSetTimeControl(
							lexical_cast<int>(parts.at(1)),
							lexical_cast<int>(parts.at(2)) * 60*100,
							lexical_cast<int>(parts.at(3)) * 100 );
					}
					catch( ... ) { }
				}
			}
			else if( readMove(m, command) ) {
				if( isUsersTurn ) {
					eventMakeMove(m);
				}
				else {
					sendOutput("It is the engine's turn to Move!");
				}
			}
			else {
				sendOutput("Can't understand this commmand");
			}
			processed = true;
			return true;
		} // CommConsole::newInput

		void CommConsole::moveAccepted() {
		}

		void CommConsole::moveIllegal() {
			sendOutput("Illegal Move!");
		}

		void CommConsole::makeMove(const Move &m) {
			if( !isUsersTurn ) {
				sendOutput(m.print());
			}
			if( isShowBoard ) {
				eventNeedCurrentState(&currentState);
				cout << currentState.print();
			}
		}

		void CommConsole::resign() {
			sendOutput("resign");
			setUsersTurn();
		}

	} // communicate
} // zochova
