
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "State.h"
#include "TranspositionTable.h"
#include "bitboardmoves.h"
#include "communicate.h"

#include "include_boost_lexical_cast.h"

using namespace std;

namespace zochova
{

	const Square State::startBoard[64] = {
		BLACK_ROOK,	 BLACK_KNIGHT, BLACK_BISHOP,BLACK_QUEEN, BLACK_KING,  BLACK_BISHOP, BLACK_KNIGHT, BLACK_ROOK,
		BLACK_PAWN,	 BLACK_PAWN,   BLACK_PAWN,	BLACK_PAWN,	 BLACK_PAWN,  BLACK_PAWN,   BLACK_PAWN,	 BLACK_PAWN,
		EMPTY,		 EMPTY,		   EMPTY,		EMPTY,		 EMPTY,		  EMPTY,		EMPTY,		 EMPTY,  
		EMPTY,		 EMPTY,		   EMPTY,		EMPTY,		 EMPTY,		  EMPTY,        EMPTY,		 EMPTY,  
		EMPTY,		 EMPTY,		   EMPTY,		EMPTY,		 EMPTY,		  EMPTY,        EMPTY,		 EMPTY,  
		EMPTY,		 EMPTY,		   EMPTY,		EMPTY,		 EMPTY,		  EMPTY,        EMPTY,		 EMPTY,  
		WHITE_PAWN,	 WHITE_PAWN,   WHITE_PAWN,	WHITE_PAWN,  WHITE_PAWN,  WHITE_PAWN,   WHITE_PAWN,   WHITE_PAWN,
		WHITE_ROOK,	 WHITE_KNIGHT, WHITE_BISHOP,WHITE_QUEEN, WHITE_KING,  WHITE_BISHOP, WHITE_KNIGHT, WHITE_ROOK
	};
/*

	const Square State::startBoard[64] = {
		BLACK_ROOK,	 EMPTY,		   EMPTY,		EMPTY,		 BLACK_KING,  EMPTY,        EMPTY,		 BLACK_ROOK,	 
		EMPTY,		 EMPTY,		   BLACK_PAWN,	EMPTY,		 EMPTY,		  EMPTY,        EMPTY,		 BLACK_PAWN,
		BLACK_PAWN,  EMPTY,		   EMPTY,		EMPTY,		 BLACK_PAWN,  EMPTY,        EMPTY,		 EMPTY,  
		EMPTY,		 BLACK_PAWN,   WHITE_QUEEN, BLACK_PAWN,  EMPTY,		  EMPTY,        BLACK_PAWN,  EMPTY,  

		WHITE_KING,  EMPTY,		   WHITE_BISHOP,EMPTY,		 WHITE_PAWN,  EMPTY,        EMPTY,		 EMPTY,  
		EMPTY,		 EMPTY,		   BLACK_QUEEN,	EMPTY,		 EMPTY,		  WHITE_PAWN,	 EMPTY,		 EMPTY,  
		WHITE_PAWN,	 EMPTY,		   WHITE_PAWN,	EMPTY,		 EMPTY,		  EMPTY,        WHITE_PAWN,	 WHITE_PAWN,	 
		EMPTY,		 EMPTY,		   EMPTY,		WHITE_ROOK,	 EMPTY,		  EMPTY,        WHITE_KNIGHT, WHITE_ROOK
	};
*/
	void State::reset() {
		memcpy(board, startBoard, sizeof(startBoard));
		next = WHITE;
		bc_l = bc_r = wc_l = wc_r = true;
		//bc_l = bc_r = true;
		//wc_l = wc_r = false;
		ep = NOPOS;
		fullmove = 1;
		halfmoveClock = 0;
		isLegalMovesActual = false;
	}

	State::State() {
		reset();
	}

	State::State(State *state) {
		(*this) = (*state);
	}

	State& State::operator=(const State &state) {
		memcpy(board, state.board, sizeof(startBoard));
		wc_l = state.wc_l;
		wc_r = state.wc_r;
		bc_l = state.bc_l;
		bc_r = state.bc_r;
		ep = state.ep;
		next = state.next;
		halfmoveClock = state.halfmoveClock;
		fullmove = state.fullmove;
		isLegalMovesActual = state.isLegalMovesActual;
		if( isLegalMovesActual ) {
			legalMoves = state.legalMoves;
		}
		return *this;
	}

	Move& Move::operator=(const Move &m) {
		memcpy(this, &m, sizeof(Move));
		return *this;
	}

	bool State::makeLegalMove(const Move &m) {
		MoveList moves = getLegalMoves();
		for( MoveList::iterator legal_m = moves.begin(); legal_m != moves.end(); legal_m++ ) {
			if( (*legal_m) == m ) {
				makeMove(m);
				return true;
			}
		}
		return false;
	}

	void State::makeMove(const Move &m) {
		if( (m.source < 0) || (m.source > 63) || (m.dest < 0) || (m.dest > 63) ) {
			throw ILLEGAL_SQUARE;
		}
		if( SQ_COLOR(board[m.source]) != next ) {
			throw ILLEGAL_MOVE;
		}
		if( SQ_COLOR(board[m.dest]) == next ) {
			throw ILLEGAL_MOVE;
		}
		if( board[m.dest] ) {
			halfmoveClock = 0;
		}
		else {
			halfmoveClock++;
		}
		if( next == BLACK ) {
			fullmove++;
		}
		if( m.dest == A1 ) {
			wc_l = false;
		}
		if( m.dest == H1 ) {
			wc_r = false;
		}
		if( m.dest == A8 ) {
			bc_l = false;
		}
		if( m.dest == H8 ) {
			bc_r = false;
		}
		switch( board[m.source] ) {
			case WHITE_KING:
				if(m.source == E1)  {
					switch(m.dest) {
						case C1: 
							board[A1] = EMPTY;
							board[D1] = WHITE_ROOK;
							break;
						case G1:
							board[H1] = EMPTY;
							board[F1] = WHITE_ROOK;
							break;
					}		
					wc_l = wc_r = false;
				}
				ep = NOPOS;
				break;
			case BLACK_KING:
				if(m.source == E8) {
					switch(m.dest) {
						case C8:
							board[A8] = EMPTY;
							board[D8] = BLACK_ROOK;
							break;
						case G8:
							board[H8] = EMPTY;
							board[F8] = BLACK_ROOK;
							break;
					}
					bc_l = bc_r = false;
				}
				ep = NOPOS;
				break;
			case WHITE_PAWN:
				halfmoveClock = 0;
				switch( m.source - m.dest ) {
					case 16:
						ep = m.dest;
						break;
					case 9:
						if( ep == m.source-1 ) {
							board[ep] = EMPTY;
						}
						ep = NOPOS;
						break;
					case 7:
						if( ep == m.source+1 ) {
							board[ep] = EMPTY;
						}
					case 8:
						ep = NOPOS;
				}
				if( (m.einfo != PROMOTE) && (m.dest < 8) ) {
					next = BLACK;
					board[m.source] = EMPTY;
					board[m.dest] = (Square)(WHITE|QUEEN);
					isLegalMovesActual = false;
				}
				if( m.einfo == PROMOTE ) {
					next = BLACK;
					board[m.source] = EMPTY;
					board[m.dest] = (Square)(WHITE|m.promote);
					isLegalMovesActual = false;
					return;
				}
				break;
			case BLACK_PAWN:
				halfmoveClock = 0;
				switch( m.dest - m.source ) {
					case 16:
						ep = m.dest;
						break;
					case 9:
						if( ep == m.source+1 ) {
							board[ep] = EMPTY;
						}
						ep = NOPOS;
						break;
					case 7:
						if( ep == m.source-1 ) {
							board[ep] = EMPTY;
						}
					case 8:
						ep = NOPOS;
				}
				if( (m.einfo != PROMOTE) && (m.dest >= 56) ) {
					next = WHITE;
					board[m.source] = EMPTY;
					board[m.dest] = (Square)(BLACK|QUEEN);
					isLegalMovesActual = false;
				}
				if( m.einfo == PROMOTE ) {
					next = WHITE;
					board[m.source] = EMPTY;
					board[m.dest] = (Square)(BLACK|m.promote);
					isLegalMovesActual = false;
					return;
				}
				break;
			case WHITE_ROOK:
				if( m.source == A1 ) {
					wc_l = false;
				}
				if( m.source == H1 ) {
					wc_r = false;
				}
				ep = NOPOS;
				break;
			case BLACK_ROOK:
				if( m.source == A8 ) {
					bc_l = false;
				}
				if( m.source == H8 ) {
					bc_r = false;
				}
				ep = NOPOS;
				break;
			default:
				ep = NOPOS;
		}
		board[m.dest] = board[m.source];
		board[m.source] = EMPTY;
		next = changeSide(next);
		isLegalMovesActual = false;
	} // State::makeMove

	string State::print() {
		string r;
		r.append("\n  ABCDEFGH  \n  -------- \n");
		int row = 8;
		for( int y = 0; y < 64; y += 8 ) {
			r.append(boost::lexical_cast<string>(row));
			r.push_back('|');
			for( int x = 0; x < 8; x++ ) {
				r.push_back(communicate::printSquare(board[y+x]));
			}
			r.push_back('|');
			r.append(boost::lexical_cast<string>(row));
			r.append("\n");
			--row;
		}
		r.append("  --------\n  ABCDEFGH \n");
		return r;
	}

	void State::genBB(
			uint64 &all_p,
			uint64 &is_my,
			uint64 &is_op,
			uint64 &is_pawn,
			uint64 &is_r,
			uint64 &is_b,
			uint64 &k_pos,
			state_castleinfo::CastleInfo &castleinfo) const
	{
		int pos;
		uint64 pos_bit = 1;
		all_p = 0;
		is_my = 0;
		is_op = 0;
		is_pawn = 0;
		is_r = 0;
		k_pos = 0;
		is_b = 0;
		castleinfo = 0;
		if( wc_r ) {
			castleinfo = state_castleinfo::SET_MY_CASTLE_RIGHT(castleinfo);
		}
		if( wc_l ) {
			castleinfo = state_castleinfo::SET_MY_CASTLE_LEFT(castleinfo);
		}
		if( bc_r ) {
			castleinfo = state_castleinfo::SET_OP_CASTLE_RIGHT(castleinfo);
		}
		if( bc_l ) {
			castleinfo =  state_castleinfo::SET_OP_CASTLE_LEFT(castleinfo);
		}
		for( pos = 0; pos < 64; pos++) {
			if( board[pos] ) {
				all_p |= pos_bit;
				if( board[pos] & next ) {
					is_my |= pos_bit;
				}
				else {
					is_op |= pos_bit;
				}
				if( board[pos] & KING ) {
					k_pos |= pos_bit;
				}
				else {
					if( board[pos] & PAWN ) {
						is_pawn |= pos_bit;
					}
					else {
						if( board[pos] & (BISHOP|QUEEN) ) {
							is_b |= pos_bit;
						}
						if( board[pos] & (ROOK|QUEEN) ) {
							is_r |= pos_bit;
						}
						//
						// 00 - knight  10 - bishop  01 - rook  11 - queen
						//
					}
				}
			}
			pos_bit <<= 1;
		}
		if( next == WHITE ) {
			all_p = FLIPVERTICAL(all_p);
			is_my = FLIPVERTICAL(is_my);
			is_op = FLIPVERTICAL(is_op);
			is_pawn = FLIPVERTICAL(is_pawn);
			is_r = FLIPVERTICAL(is_r);
			k_pos = FLIPVERTICAL(k_pos);
			is_b = FLIPVERTICAL(is_b);
			castleinfo = state_castleinfo::FLIP(castleinfo);
		}
	} // State::genBB

	MoveList State::getLegalMoves() {
		if( isLegalMovesActual ) {
			return legalMoves;
		}
		legalMoves.clear();
		uint64 intlist[128], templist[128];
		uint64 *intm;
		Move m;
		uint64 all_p, is_my, is_op, is_pawn, is_r, is_b, k_pos;
		uint64 tall_p, tis_my, tis_op, tis_pawn, tis_r, tis_b, tk_pos;
		state_castleinfo::CastleInfo cinfo, tcinfo;
		bitboardmoves::initialize();
		genBB(all_p, is_my, is_op, is_pawn, is_r, is_b, k_pos, cinfo);
		bitboardmoves::gen(all_p, is_my, is_op, is_pawn, is_r, is_b, k_pos, intlist);
		intm = intlist;
		while( *intm ) {
			GET_LS1B_INDEX((unsigned long*)(&(m.source)), (*intm)&is_my);
			GET_LS1B_INDEX((unsigned long*)(&(m.dest)), (*intm)&(~is_my));
			if( next == WHITE ) {
				m.source = (BoardPosition)(8*(7-m.source/8)+m.source%8);
				m.dest = (BoardPosition)(8*(7-m.dest/8)+m.dest%8);
			}
			m.einfo = NONE;
			State *tempState = new State(this);
			tempState->makeMove(m);
			tempState->genBB(tall_p, tis_my, tis_op, tis_pawn, tis_r, tis_b, tk_pos, tcinfo);
			if( (tk_pos & tis_op)
				&& ((tk_pos & bitboardmoves::gen(tall_p, tis_my, tis_op, tis_pawn, tis_r, tis_b, tk_pos, templist)) == 0) )
			{
				legalMoves.push_back(m);
			}
			delete tempState;
			intm++;
		}

		return legalMoves;
	} // State::getLegalMoves

	std::string State::toFen() const {
		std::string fen;
		int pos = 0;
		unsigned char n_EMPTY = 0;
		do {
			if( (pos > 0) && ((pos%8) == 0) ) {
				if( n_EMPTY ) {
					fen.push_back('0'+n_EMPTY);
					n_EMPTY = 0;
				}
				fen.push_back('/');
			}
			if( board[pos] ) {
				if( n_EMPTY ) {
					fen.push_back('0'+n_EMPTY);
					n_EMPTY = 0;
				}
				fen.push_back(communicate::printSquare(board[pos]));
			}
			else {
				n_EMPTY++;
			}
			pos++;
		} while( pos < 64 );
		fen.push_back(' ');
		if( next == WHITE ) {
			fen.push_back('w');
		}
		else {
			fen.push_back('b');
		}
		fen.push_back(' ');
		if( bc_l | bc_r | wc_l | wc_r ) {
			if( wc_r ) {
				fen.push_back('K');
			}
			if( wc_l ) {
				fen.push_back('Q');
			}
			if( bc_r ) {
				fen.push_back('k');
			}
			if( bc_l ) {
				fen.push_back('q');
			}
		}
		else {
			fen.push_back('-');
		}
		fen.push_back(' ');
		if( ep != NOPOS ) {
			fen.push_back(ep%8+'a');
			fen.push_back(7-(ep/8)+'1');
		}
		else {
			fen.push_back('-');
		}
		fen.push_back(' ');
		fen.append(boost::lexical_cast<string>(halfmoveClock));
		fen.push_back(' ');
		fen.append(boost::lexical_cast<string>(fullmove));
		return fen;
	} // State::toFen

	string Move::print() const{
		string r;
		r.push_back(source%8+'a');
		r.push_back(7-(source/8)+'1');
		r.push_back(dest%8+'a');
		r.push_back(7-(dest/8)+'1');
		if( einfo == PROMOTE ) {
			r.push_back(communicate::printSquare((Square)(BLACK|promote)));
		}
		return r;
	}

	bool Move::operator==(const Move &m) {
		if( source == m.source && dest == m.dest && einfo == m.einfo ) {
			if( einfo != NONE ) {
				return promote == m.promote;
			}
			return true;
		}
		return false;
	}


	string Move::toAlgebraicNotation(const State &st) const {
		string r;
		// TODO this shit
		return r;
	}

	const Square State::operator[](const BoardPosition pos) const {
		return board[pos];
	}

	const BoardPosition State::getEnPassantPosition() const {
		return ep;
	}

	string printMoveList(const MoveList& ml) {
		string ret;
		for( MoveList::const_iterator m = ml.begin(); m != ml.end(); m++ ) {
			ret.append( m->print() );
			ret.append( " " );
		}
		return ret;
	}

	uint64 Move::toBitBoard() {
		switch( einfo ) {
			case NONE:
				return (U64(1)<<source) | (U64(1)<<dest);
			case CASTLE_LEFT:
				return CL_SQUARES;
			case CASTLE_RIGHT:
				return CR_SQUARES;
			case PROMOTE:
				switch( this->promote ) {
					case QUEEN:
						return M_PROMOTE_Q | (U64(1)<<source) | (U64(1)<<dest);
					case KNIGHT:
						return M_PROMOTE_K | (U64(1)<<source) | (U64(1)<<dest);
					case BISHOP:
						return M_PROMOTE_B | (U64(1)<<source) | (U64(1)<<dest);
					case ROOK:
						return M_PROMOTE_R | (U64(1)<<source) | (U64(1)<<dest);
				}
			case EN_PASSANT:
				return M_ENP | (U64(1)<<source) | (U64(1)<<dest);
			default:
				return 0;
		}
	}

	uint64 Move::toBitBoard(int s,
								   int d,
								   uint64 is_my,
								   uint64 is_op,
								   uint64 k_pos,
								   uint64 is_pawn)
	{
		uint64 s64 = U64(1)<<s;
		uint64 d64 = U64(1)<<d;
		if( (s64 == SQ_E1) && (s64 & k_pos) ) {
			if( d64 == SQ_G1 ) {
				return CR_SQUARES;
			}
			if( d64 == SQ_C1 ) {
				return CL_SQUARES;
			}
		}
		if( s64 & is_pawn) {
			if( s64 & OPPONENT_LINES ) { // promotion
				return M_PROMOTE_Q | s64|d64;
			}
			if( d64 & (s64<<7 | s64<<9) & ~is_op ) { // en passant
				return M_ENP | s64|d64;
			}
		}
		return s64|d64;
	}

	void Move::fromBitBoard(uint64 m, uint64 my) {
		if( m == CR_SQUARES ) {
			m = SQ_E1|SQ_G1;
			einfo = CASTLE_RIGHT;
		}
		else {
			if( m == CL_SQUARES ) {
				m = SQ_E1|SQ_C1;
				einfo = CASTLE_LEFT;
			}
			else {
				switch( m & M_PROMOTE_ALL ) {
					case M_PROMOTE_Q:
						einfo = PROMOTE;
						promote = QUEEN;
						m &= ~M_PROMOTE_ALL;
						break;
					case M_PROMOTE_B:
						einfo = PROMOTE;
						promote = BISHOP;
						m &= ~M_PROMOTE_ALL;
						break;
					case M_PROMOTE_K:
						einfo = PROMOTE;
						promote = KNIGHT;
						m &= ~M_PROMOTE_ALL;
						break;
					case M_PROMOTE_R:
						einfo = PROMOTE;
						promote = ROOK;
						m &= ~M_PROMOTE_ALL;
						break;
					default:
						if( (m & M_ENP) == M_ENP ) {
							m = m&~M_ENP;
							einfo = EN_PASSANT;
						}
						else {
							einfo = NONE;
						}
				}
			}
		}
		source = (BoardPosition)LS1B_INDEX(m&my);
		dest = (BoardPosition)LS1B_INDEX(m&(~my));
	}

	bool State::isMate() {
		// TODO
		return false;
	}

	bool State::isStalemate() {
		// TODO
		return false;
	}

	const int State::getPieceCount() const {
		const Square* square;
		int pieceCount;
		pieceCount = 0;
		square = &(board[63]);
		do {
			if( (*square) != EMPTY ) {
				++pieceCount;
			}
			--square;
		} while( square != board );
		return pieceCount;
	}

	const bool State::canBlackCastleLeft() const {
		return bc_l;	
	}

	const bool State::canBlackCastleRight() const {
		return bc_r;
	}

	const bool State::canWhiteCastleLeft() const {
		return wc_l;
	}

	const bool State::canWhiteCastleRight() const {
		return wc_r;
	}


} // zochova
