
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "search_flags.h"
#include "Searcher.h"
#include "TranspositionTable.h"
#include "evaluation.h"
#include "Thinker.h"
#if ENGINE_LOG_LEVEL > 1
#	include "log.h"
#endif


namespace zochova
{
	namespace engine
	{
		namespace search
		{


			void Searcher::_qsearch_(
				BB_ARGUMENTS,
			    int dynamicValue,
				int &alpha,
				int &beta,
				int &bestReturned,
			    uint64 &nodeCount,
				uint64 s,
				uint64 d,
				int qply)
			{
				using namespace evaluation;

				int val;

				assert(is_op & d );

				if( (d & CENTER_SQ) && (s & (~CENTER_SQ) & (~(k_pos/*|is_pawn*/))) ) {
					dynamicValue += CENTER_SQ_VALUE;
				}
				else {
					if( (d & (~CENTER_SQ)) && (s & CENTER_SQ & (~(k_pos/*|is_pawn*/))) ) {
						dynamicValue -= CENTER_SQ_VALUE;
					}
				}
				if( is_pawn & d ) {
					dynamicValue += PAWN_VALUE;
					if( d & ROW_2 ) {
						dynamicValue += PAWN_RANK7_VALUE;
					}
					else {
						if( d & ROW_3 ) {
							dynamicValue += PAWN_RANK6_VALUE;
						}
						else {
							if( d & ROW_4 ) {
								dynamicValue += PAWN_RANK5_VALUE;
							}
						}
					}
					is_pawn ^= d;
				}
				else {
					if( is_r & d ) {
						is_r ^= d;
						if( is_b & d ) {
							is_b ^= d;
							dynamicValue += QUEEN_VALUE;
						}
						else {
							dynamicValue += ROOK_VALUE;
						}
					}
					else {
						if( is_b & d ) {
							is_b ^= d;
							dynamicValue += BISHOP_VALUE;
						}
						else {
							dynamicValue += KNIGHT_VALUE;
						}
					}
					if( d & CENTER_SQ ) {
						dynamicValue += CENTER_SQ_VALUE;
					}
				}

				val = -_qsearch(
					FLIPVERTICAL(all_p ^ s),
					FLIPVERTICAL(is_op ^ d),
					FLIPVERTICAL(is_my ^ (s|d)),
					FLIPVERTICAL((is_pawn&s) ? (is_pawn ^ (s|d)) : is_pawn ),
					FLIPVERTICAL((is_r&s) ? (is_r ^ (s|d)) : is_r),
					FLIPVERTICAL((is_b&s) ? (is_b ^ (s|d)) : is_b),
					FLIPVERTICAL((k_pos&s) ? (k_pos ^ (s|d)) : k_pos),
					-dynamicValue, 
					-beta,
					-alpha,
					nodeCount,
					qply-1);
				if( val != SEARCH_ILLEGAL_MOVE_VALUE ) {
					if( val > bestReturned ) {
						bestReturned = val;
					}
					if( val > alpha ) {
						alpha = val;
					}
				}
			}

			void __fastcall Searcher::_q_ray_attack_capture(
					   const uint64 DIR_EDGE,
					   const int SHIFT,
					   BB_ARGUMENTS,
					   const uint64 pos_bit,
					   int &alpha,
					   int &beta,
					   int &bestReturned,
					   int dynamicValue,
					   uint64 &nodeCount,
					   int qply)
			{
				uint64 d;
				d = pos_bit;
				while( !(d & (DIR_EDGE)) ) {
					d = ROR(d, SHIFT);
					if( d & all_p ) {
						if( d & is_op ) {
							_qsearch_(
								BB_ARGUMENTS_CALL,
								dynamicValue,
								alpha, beta, 
								bestReturned,
								nodeCount, 
								pos_bit,
								d,
								qply);
						}
						break;
					}
				}
			}


			int Searcher::_qsearch(
				BB_ARGUMENTS,
			    int dynamicValue,
				int alpha,
				int beta,
			    uint64 &nodeCount,
				int qply)
			{

				using namespace evaluation;
				using namespace bitboardmoves;

				int bestReturned;
				bestReturned = -INFINITE_VALUE;

				if( kingChecked(
						all_p, is_op, is_my, is_pawn, is_r, is_b, k_pos, true) )
				{
					return -SEARCH_ILLEGAL_MOVE_VALUE;
				}

				++nodeCount;
/*
				if( (alpha >= dynamicValue + QUEEN_VALUE + PAWN_VALUE) && (qply < 4) ) {
					return dynamicValue;
				}


				if( dynamicValue < alpha - QUEEN_VALUE ) {
					return dynamicValue;
				}
*/
				if( dynamicValue > alpha ) {
					alpha = dynamicValue;
					if( dynamicValue >= beta ) {
						return dynamicValue;
					}
				}

				uint64 tmy, pos_bit;
				tmy = is_my;
				int tval;

				do { 
					uint64 d;
					pos_bit = LS1B(tmy);
					tval = dynamicValue;
					if( is_pawn & pos_bit ) {
						if( pos_bit&OPPONENT_LINES ) {
							if( (pos_bit<<8 & all_p) == 0 ) {
								tval = _qsearch(
									FLIPVERTICAL(all_p ^ (pos_bit | (pos_bit<<8))),
									FLIPVERTICAL(is_op),
									FLIPVERTICAL(is_my ^ (pos_bit | (pos_bit<<8))),
									FLIPVERTICAL(is_pawn ^ pos_bit),
									FLIPVERTICAL(is_r | (pos_bit<<8)),
									FLIPVERTICAL(is_b | (pos_bit<<8)),
									FLIPVERTICAL(k_pos),
									dynamicValue - PAWN_VALUE - PAWN_RANK7_VALUE + QUEEN_VALUE,
									-beta,
									-alpha,
									nodeCount,
									qply-1);
								if( tval > bestReturned ) {
									bestReturned = tval;
									if( tval > alpha ) {
										alpha = tval;
										if( alpha >= beta ) {
											return alpha;
										}
									}
								}
							}
							if( (!(pos_bit&EDGE_RIGHT)) && (is_op & (pos_bit<<9)) ) {
						/*		(*list) = pos_bit | (pos_bit<<9) | M_PROMOTE_Q;
								++list;
								(*list) = pos_bit | (pos_bit<<9) | M_PROMOTE_K;
								++list;
								(*list) = pos_bit | (pos_bit<<9) | M_PROMOTE_R;
								++list;
								(*list) = pos_bit | (pos_bit<<9) | M_PROMOTE_B;
								++list;
								r |= (pos_bit<<9);*/
							}
							if( (!(pos_bit&EDGE_LEFT)) && (is_op & (pos_bit<<7)) ) {
							/*	(*list) = pos_bit | (pos_bit<<7) | M_PROMOTE_Q;
								++list;
								(*list) = pos_bit | (pos_bit<<7) | M_PROMOTE_K;
								++list;
								(*list) = pos_bit | (pos_bit<<7) | M_PROMOTE_R;
								++list;
								(*list) = pos_bit | (pos_bit<<7) | M_PROMOTE_B;
								++list;
								r |= (pos_bit<<7);*/
							}
							continue;
						}
						else {
							if( pos_bit & (ROW_4|ROW_5|ROW_6) ) {
								++tval;
							}
							if( (!(pos_bit&EDGE_RIGHT)) && (is_op & (pos_bit<<9)) ) {
								_qsearch_(
									BB_ARGUMENTS_CALL,
									tval,
									alpha, beta, 
									bestReturned,
									nodeCount, 
									pos_bit,
									pos_bit<<9,
									qply);
								if( alpha >= beta ) {
									return alpha;
								}

							}
							if( (!(pos_bit&EDGE_LEFT)) && (is_op & (pos_bit<<7)) ) {
								_qsearch_(
									BB_ARGUMENTS_CALL,
									tval,
									alpha, beta,
									bestReturned,
									nodeCount, 
									pos_bit,
									pos_bit<<7,
									qply);
								if( alpha >= beta ) {
									return alpha;
								}
							}
						}
						continue;
					}

					if( is_b&pos_bit ) {
						_q_ray_attack_capture(
							EDGE_RIGHT|EDGE_DOWN, 64-9, 
							BB_ARGUMENTS_CALL, pos_bit, 
							alpha, beta, bestReturned, tval, nodeCount, qply);
						if( alpha >= beta ) {
							return alpha;
						}
						_q_ray_attack_capture(
							EDGE_LEFT|EDGE_DOWN, 64-7, 
							BB_ARGUMENTS_CALL, pos_bit, 
							alpha, beta, bestReturned, tval, nodeCount, qply);
						if( alpha >= beta ) {
							return alpha;
						}
						_q_ray_attack_capture(
							EDGE_LEFT|EDGE_UP, 9, 
							BB_ARGUMENTS_CALL, pos_bit, 
							alpha, beta, bestReturned, tval, nodeCount, qply);
						if( alpha >= beta ) {
							return alpha;
						}
						_q_ray_attack_capture(
							EDGE_RIGHT|EDGE_UP, 7, 
							BB_ARGUMENTS_CALL, pos_bit, 
							alpha, beta, bestReturned, tval, nodeCount, qply);
						if( alpha >= beta ) {
							return alpha;
						}
					}
					if( is_r&pos_bit ) {
						_q_ray_attack_capture(
							EDGE_LEFT, 1, 
							BB_ARGUMENTS_CALL, pos_bit, 
							alpha, beta, bestReturned, tval, nodeCount, qply);
						if( alpha >= beta ) {
							return alpha;
						}
						_q_ray_attack_capture(
							EDGE_RIGHT, 64-1, 
							BB_ARGUMENTS_CALL, pos_bit, 
							alpha, beta, bestReturned, tval, nodeCount, qply);
						if( alpha >= beta ) {
							return alpha;
						}
						_q_ray_attack_capture(
							EDGE_UP, 8, 
							BB_ARGUMENTS_CALL, pos_bit, 
							alpha, beta, bestReturned, tval, nodeCount, qply);
						if( alpha >= beta ) {
							return alpha;
						}
						_q_ray_attack_capture(
							EDGE_DOWN, 64-8, 
							BB_ARGUMENTS_CALL, pos_bit,
							alpha, beta, bestReturned, tval, nodeCount, qply);
						if( alpha >= beta ) {
							return alpha;
						}
						continue;
					}
					if( pos_bit & k_pos ) {
						d = king_moves[LS1B_INDEX(pos_bit)] & is_op;
						if( d != 0 ) {
							do {
								if( (king_moves[LS1B_INDEX(d)] & is_op & k_pos) == 0 ) {
									_qsearch_(
										BB_ARGUMENTS_CALL,
										tval,
										alpha, beta, 
										bestReturned, 
										nodeCount,
										pos_bit,
										LS1B(d),
										qply);
									if( alpha >= beta ) {
										return alpha;
									}
								}
							} while( d -= LS1B(d) );
						}
						continue;
					}
					if( !(pos_bit & (is_pawn|is_r|is_b|k_pos)) ) {
						d = knight_moves[LS1B_INDEX(pos_bit)] & is_op;
						if( d != 0 ) {
							do {
								_qsearch_(
									BB_ARGUMENTS_CALL,
									tval,
									alpha, beta, 
									bestReturned, 
									nodeCount,
									pos_bit,
									LS1B(d),
									qply);
								if( alpha >= beta ) {
									return alpha;
								}
							} while( d -= LS1B(d) );
						}
					}
				} while( tmy -= pos_bit );

				if( bestReturned > dynamicValue )  {
					return bestReturned;
				}
				else {
					return dynamicValue;
				}



			} // Searcher::_qsearch

		} // search
	} // engine
} // zochova
