
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "bitboardmoves.h"
#include "State.h"
#include "TranspositionTable.h"

namespace zochova
{
	namespace bitboardmoves
	{

		uint64 knight_moves[64];
		uint64 king_moves[64];

		void initialize() {
			static bool inited = false;
			uint64* pm;
			uint64 pos_bit;
			if( inited ) {
				return;
			}
			int pos;
			pos = 64;
			pos_bit = U64(1)<<63;
			do {
				int file, rank;
				pos--;

				file = pos%8;
				rank = pos/8;

				// knights moves
				pm = &(knight_moves[pos]);
				(*pm) = 0;
				if( (rank < 7) && (file < 6) ) {
					(*pm) |= pos_bit<<10;
				}
				if( (rank < 7) && (file > 1) ) {
					(*pm) |= pos_bit<<6;
				}
				if( (rank < 6) && (file < 7) ) {
					(*pm) |= pos_bit<<17;
				}
				if( (rank < 6) && (file > 0) ) {
					(*pm) |= pos_bit<<15;
				}
				if( (rank > 0) && (file > 1) ) {
					(*pm) |= pos_bit>>10;
				}
				if( (rank > 0) && (file < 6) ) {
					(*pm) |= pos_bit>>6;
				}
				if( (rank > 1) && (file > 0) ) {
					(*pm) |= pos_bit>>17;
				}
				if( (rank > 1) && (file < 7) ) {
					(*pm) |= pos_bit>>15;
				}

				// kings moves
				pm = &(king_moves[pos]);
				(*pm) = 0;
				if( file < 7 ) {
					(*pm) |= pos_bit<<1;
				}
				if( rank < 7 ) {
					(*pm) |= pos_bit<<8;
				}
				if( (rank < 7) && (file > 0) ) {
					(*pm) |= pos_bit<<7;
				}
				if( (rank < 7) && (file < 7) ) {
					(*pm) |= pos_bit<<9;
				}
				if( file > 0 ) {
					(*pm) |= pos_bit>>1;
				}
				if( rank > 0 ) {
					(*pm) |= pos_bit>>8;
				}
				if( (rank > 0) && (file < 7) ) {
					(*pm) |= pos_bit>>7;
				}
				if( (rank > 0) && (file > 0) ) {
					(*pm) |= pos_bit>>9;
				}
				pos_bit >>= 1;
			} while( pos );
			inited = true;
		} // initialize

		void diag_attacks(const uint64 x,
						  const int index,
						  uint64 &sw,
						  uint64 &se,
						  uint64 &nw,
						  uint64 &ne)
		{
			if( x&DIAG0_HALF1 ) {
				sw = (DIAG0<<index)&DIAG0_HALF1;
				ne = (DIAG0>>(63-index))&DIAG0_HALF1;
			}
			else {
				sw = (DIAG0<<index)&DIAG0_HALF0;
				ne = (DIAG0>>(63-index))&DIAG0_HALF0;
			}
			if( x&DIAG1_HALF1 ) {
				nw = (DIAG1>>(63-index))&DIAG1_HALF1-x;
				se = (DIAG1<<index)&DIAG1_HALF1-x;
			}
			else {
				nw = (DIAG1>>(63-index))&DIAG1_HALF0-x;
				se = (DIAG1<<index)&DIAG1_HALF0-x;
			}
		}

		void horver_attacks(const uint64 x,
							const int index,
							uint64 &s,
							uint64 &e,
							uint64 &n,
							uint64 &w)
		{
			s = EDGE_LEFT << (index%8);
			n = s&(x-1);
			s &= ~(x|(x-1));
			w = EDGE_UP << (8*(index/8));
			e = w & (w << (index%8+1));
			w &= w >> (8-(index%8));
		}

		void __fastcall ray_attack(
				   const uint64 DIR_EDGE,
				   const int SHIFT,
				   const uint64 pos_bit,
				   const uint64 all_p,
				   const uint64 is_op,
				   uint64 &r,
				   uint64* &list)
		{
			uint64 d;
			d = pos_bit;
			while( !(d & (DIR_EDGE)) ) {
				d = ROR(d, SHIFT);
				if( d & all_p ) {
					if( d & is_op ) {
						(*list) = pos_bit|d;
						r |= d;
						list++;
					}
					break;
				}
				(*list) = pos_bit|d;
				r |= d;
				list++;
			}
		}

		void __fastcall ray_attack_capture(
				   const uint64 DIR_EDGE,
				   const int SHIFT,
				   const uint64 pos_bit,
				   const uint64 all_p,
				   const uint64 is_op,
				   uint64 &r,
				   uint64* &list)
		{
			uint64 d;
			d = pos_bit;
			while( !(d & (DIR_EDGE)) ) {
				d = ROR(d, SHIFT);
				if( d & all_p ) {
					if( d & is_op ) {
						(*list) = pos_bit|d;
						r |= d;
						list++;
					}
					break;
				}
			}
		}

		int __fastcall ray_attack_mobility(
				   const uint64 DIR_EDGE,
				   const int SHIFT,
				   const uint64 pos_bit,
				   const uint64 all_p,
				   const uint64 is_op,
				   uint64 &des)
		{
			uint64 d;
			int r;
			r = 0;
			d = pos_bit;
			while( !(d & (DIR_EDGE)) ) {
				d = ROR(d, SHIFT);
				if( d & all_p ) {
					if( d & is_op ) {
						++r;
						des |= d;
					}
					break;
				}
				++r;
				des |= d;
			}
			return r;
		}

		uint64 gen(BB_ARGUMENTS, uint64 *list, uint64 **listEnd, bool plausibles) {
			uint64 tmy, pos_bit, r;
			r = 0;
			tmy = is_my;

			do { 
				uint64 d;
				pos_bit = LS1B(tmy);
				if( is_pawn & pos_bit ) {
					if( pos_bit&OPPONENT_LINES ) {
						if( !plausibles ) {
						if( (pos_bit<<8 & all_p) == 0 ) {
							(*list) = pos_bit | (pos_bit<<8) | M_PROMOTE_Q;
							++list;
							(*list) = pos_bit | (pos_bit<<8) | M_PROMOTE_K;
							++list;
							(*list) = pos_bit | (pos_bit<<8) | M_PROMOTE_R;
							++list;
							(*list) = pos_bit | (pos_bit<<8) | M_PROMOTE_B;
							++list;
							r |= (pos_bit<<8);
						}
						if( (!(pos_bit&EDGE_RIGHT)) && (is_op & (pos_bit<<9)) ) {
							(*list) = pos_bit | (pos_bit<<9) | M_PROMOTE_Q;
							++list;
							(*list) = pos_bit | (pos_bit<<9) | M_PROMOTE_K;
							++list;
							(*list) = pos_bit | (pos_bit<<9) | M_PROMOTE_R;
							++list;
							(*list) = pos_bit | (pos_bit<<9) | M_PROMOTE_B;
							++list;
							r |= (pos_bit<<9);
						}
						if( (!(pos_bit&EDGE_LEFT)) && (is_op & (pos_bit<<7)) ) {
							(*list) = pos_bit | (pos_bit<<7) | M_PROMOTE_Q;
							++list;
							(*list) = pos_bit | (pos_bit<<7) | M_PROMOTE_K;
							++list;
							(*list) = pos_bit | (pos_bit<<7) | M_PROMOTE_R;
							++list;
							(*list) = pos_bit | (pos_bit<<7) | M_PROMOTE_B;
							++list;
							r |= (pos_bit<<7);
						}
						}
						continue;
					}
					else {
						if( (!(pos_bit&EDGE_RIGHT)) && (is_op & (pos_bit<<9)) ) {
							(*list) = pos_bit|(pos_bit<<9);
							r |= (pos_bit<<9);
							list++;
						}
						if( (!(pos_bit&EDGE_LEFT)) && (is_op & (pos_bit<<7)) ) {
							(*list) = pos_bit|(pos_bit<<7);
							r |= (pos_bit<<7);
							list++;
						}
						if( !plausibles ) {
							if( !(all_p & (pos_bit<<8)) ) {
								(*list) = pos_bit|(pos_bit<<8);
								r |= (pos_bit<<8);
								list++;
								if( (pos_bit&SECOND_ROW) && !(all_p & (pos_bit<<16)) ) {
									(*list) = pos_bit|(pos_bit<<16);
									r |= (pos_bit<<16);
									list++;
								}
							}
						}
					}
					continue;
				}

				if( plausibles ) { 
					if( is_b&pos_bit ) {
						ray_attack_capture(EDGE_RIGHT|EDGE_DOWN, 64-9, pos_bit, all_p, is_op, r, list);
						ray_attack_capture(EDGE_LEFT|EDGE_DOWN, 64-7, pos_bit, all_p, is_op, r, list);
						ray_attack_capture(EDGE_LEFT|EDGE_UP, 9, pos_bit, all_p, is_op, r, list);
						ray_attack_capture(EDGE_RIGHT|EDGE_UP, 7, pos_bit, all_p, is_op, r, list);
					}
					if( is_r&pos_bit ) {
						ray_attack_capture(EDGE_LEFT, 1, pos_bit, all_p, is_op, r, list);
						ray_attack_capture(EDGE_RIGHT, 64-1, pos_bit, all_p, is_op, r, list);
						ray_attack_capture(EDGE_UP, 8, pos_bit, all_p, is_op, r, list);
						ray_attack_capture(EDGE_DOWN, 64-8, pos_bit, all_p, is_op, r, list);
						continue;
					}
				}
				else {
					if( is_b&pos_bit ) {
						ray_attack(EDGE_RIGHT|EDGE_DOWN, 64-9, pos_bit, all_p, is_op, r, list);
						ray_attack(EDGE_LEFT|EDGE_DOWN, 64-7, pos_bit, all_p, is_op, r, list);
						ray_attack(EDGE_LEFT|EDGE_UP, 9, pos_bit, all_p, is_op, r, list);
						ray_attack(EDGE_RIGHT|EDGE_UP, 7, pos_bit, all_p, is_op, r, list);
					}
					if( is_r&pos_bit ) {
						ray_attack(EDGE_LEFT, 1, pos_bit, all_p, is_op, r, list);
						ray_attack(EDGE_RIGHT, 64-1, pos_bit, all_p, is_op, r, list);
						ray_attack(EDGE_UP, 8, pos_bit, all_p, is_op, r, list);
						ray_attack(EDGE_DOWN, 64-8, pos_bit, all_p, is_op, r, list);
						continue;
					}
				}
				if( pos_bit & k_pos ) {
					if( plausibles ) {
						d = king_moves[LS1B_INDEX(pos_bit)] & is_op;
					}
					else {
						d = king_moves[LS1B_INDEX(pos_bit)] & (~is_my);
					}
					if( d != 0 ) {
						do {
							if( (king_moves[LS1B_INDEX(d)] & is_op & k_pos) == 0 ) {
								r |= LS1B(d);
								(*list) = pos_bit|LS1B(d);
								list++;
							}
						} while( d -= LS1B(d) );
					}
					continue;
				}
				if( !(pos_bit & (is_pawn|is_r|is_b|k_pos)) ) {
					if( plausibles ) {
						d = knight_moves[LS1B_INDEX(pos_bit)] & is_op;
					}
					else {
						d = knight_moves[LS1B_INDEX(pos_bit)] & (~is_my);
					}
					r |= d;
					if( d != 0 ) {
						do {
							(*list) = pos_bit|LS1B(d);
							list++;
						} while( d -= LS1B(d) );
					}
				}
				if( r & k_pos ) {
					return r;
				}
			} while( tmy -= pos_bit );
			(*list) = 0;
			if( listEnd != 0 ) {
				(*listEnd) = list;
			}
			return r;

		} // gen

		int mobility(BB_ARGUMENTS) {
			uint64 tmy, pos_bit;
			tmy = is_my;
			int r;
			r = 0;

			do { 
				uint64 d, des;
				pos_bit = LS1B(tmy);
				des = 0;
				if( is_pawn & pos_bit ) {
					if( (!(pos_bit&EDGE_RIGHT)) && (is_op & (pos_bit<<9)) ) {
						++r;
						des |= pos_bit<<9;
					}
					if( (!(pos_bit&EDGE_LEFT)) && (is_op & (pos_bit<<7)) ) {
						++r;
						des |= pos_bit<<7;
					}
					if( !(all_p & (pos_bit<<8)) ) {
						++r;
						des |= pos_bit<<8;
						if( (pos_bit&SECOND_ROW) && !(all_p & (pos_bit<<16)) ) {
							des |= pos_bit<<16;
							++r;
						}
					}
					//continue;
				}
				else {
					if( is_b&pos_bit ) {
						r += ray_attack_mobility(EDGE_RIGHT|EDGE_DOWN, 64-9, pos_bit, all_p, is_op, des);
						r += ray_attack_mobility(EDGE_LEFT|EDGE_DOWN, 64-7, pos_bit, all_p, is_op, des);
						r += ray_attack_mobility(EDGE_LEFT|EDGE_UP, 9, pos_bit, all_p, is_op, des);
						r += ray_attack_mobility(EDGE_RIGHT|EDGE_UP, 7, pos_bit, all_p, is_op, des);
					}
					if( is_r&pos_bit ) {
						r += ray_attack_mobility(EDGE_LEFT, 1, pos_bit, all_p, is_op, des);
						r += ray_attack_mobility(EDGE_RIGHT, 64-1, pos_bit, all_p, is_op, des);
						r += ray_attack_mobility(EDGE_UP, 8, pos_bit, all_p, is_op, des);
						r += ray_attack_mobility(EDGE_DOWN, 64-8, pos_bit, all_p, is_op, des);
						//continue;
					}
					else {
	/*				if( pos_bit & k_pos ) {
						d = king_moves[LS1B_INDEX(pos_bit)] & (~is_my);
						if( d != 0 ) {
							do {
								if( !(king_moves[LS1B_INDEX(d)] & is_op & k_pos ) ) {
									++r;
								}
							} while( d -= LS1B(d) );
						}
						continue;
					}*/
						if( !(pos_bit & (is_pawn|is_r|is_b|k_pos)) ) {
							r += POPCOUNT(knight_moves[LS1B_INDEX(pos_bit)] & (~is_my));
							des |= knight_moves[LS1B_INDEX(pos_bit)] & (~is_my);
						}
					}
				}
				r += POPCOUNT(des & CENTER_SQ);
			} while( tmy -= pos_bit );
			return r;
		} // mobility

		uint64 genCaptures(BB_ARGUMENTS, uint64 *list) {
			uint64 tmy, pos_bit, r;
			r = 0;
			tmy = is_my;

			do { 
				uint64 d;
				pos_bit = LS1B(tmy);
				if( is_pawn & pos_bit ) {
					if( pos_bit&OPPONENT_LINES ) {
						continue;
					}
					if( (!(pos_bit&EDGE_RIGHT)) && (is_op & (pos_bit<<9)) ) {
						(*list) = pos_bit|(pos_bit<<9);
						r |= (pos_bit<<9);
						list++;
					}
					if( (!(pos_bit&EDGE_LEFT)) && (is_op & (pos_bit<<7)) ) {
						(*list) = pos_bit|(pos_bit<<7);
						r |= (pos_bit<<7);
						list++;
					}
					continue;
				}

				if( is_b&pos_bit ) {
					ray_attack_capture(EDGE_RIGHT|EDGE_DOWN, 64-9, pos_bit, all_p, is_op, r, list);
					ray_attack_capture(EDGE_LEFT|EDGE_DOWN, 64-7, pos_bit, all_p, is_op, r, list);
					ray_attack_capture(EDGE_LEFT|EDGE_UP, 9, pos_bit, all_p, is_op, r, list);
					ray_attack_capture(EDGE_RIGHT|EDGE_UP, 7, pos_bit, all_p, is_op, r, list);
				}
				if( is_r&pos_bit ) {
					ray_attack_capture(EDGE_LEFT, 1, pos_bit, all_p, is_op, r, list);
					ray_attack_capture(EDGE_RIGHT, 64-1, pos_bit, all_p, is_op, r, list);
					ray_attack_capture(EDGE_UP, 8, pos_bit, all_p, is_op, r, list);
					ray_attack_capture(EDGE_DOWN, 64-8, pos_bit, all_p, is_op, r, list);
					continue;
				}

				if( pos_bit & k_pos ) {
					d = king_moves[LS1B_INDEX(pos_bit)] & is_op;
					if( d != 0 ) {
						do {
							if( !(king_moves[LS1B_INDEX(d)] & is_op & k_pos ) ) {
								r |= LS1B(d);
								(*list) = pos_bit|LS1B(d);
								list++;
							}
						} while( d -= LS1B(d) );
					}
					continue;
				}

				if( !(pos_bit & (is_pawn|is_r|is_b|k_pos)) ) {
					d = knight_moves[LS1B_INDEX(pos_bit)] & is_op;
					if( d != 0 ) {
						r |= d;
						do {
							(*list) = pos_bit|LS1B(d);
							list++;
						} while( d -= LS1B(d) );
					}
				}

			} while( tmy -= pos_bit );
			(*list) = 0;

			return r;

		} // genCaptures

		bool __fastcall king_ray_attack(
					   const uint64 DIR_EDGE,
					   const int SHIFT,
					   const uint64 k,
					   const uint64 all_p,
					   const uint64 is_op)
		{
			uint64 d;
			d = k;
			while( !(d & (DIR_EDGE)) ) {
				d = ROR(d, SHIFT);
				if( d & all_p ) {
					if( d & is_op ) {
						return true;
					}
					return false;
				}
			}
			return false;
		}

		bool kingChecked(BB_ARGUMENTS, bool isFlipped) {
			uint64 k;
			k = k_pos&is_my;

			assert( (king_moves[LS1B_INDEX(k)] & is_op & k_pos) == 0);
			
			if( isFlipped ) {
				if( (!(k&EDGE_RIGHT)) && (is_op & (k>>7) & is_pawn) ) {
					return true;
				}
				if( (!(k&EDGE_LEFT)) && (is_op & (k>>9) & is_pawn) ) {
					return true;
				}
			}
			else {
				if( (!(k&EDGE_RIGHT)) && (is_op & (k<<9) & is_pawn) ) {
					return true;
				}
				if( (!(k&EDGE_LEFT)) && (is_op & (k<<7) & is_pawn) ) {
					return true;
				}
			}
			if( knight_moves[LS1B_INDEX(k)] & is_op & (~(is_pawn|is_b|is_r|k_pos)) ) {
				return true;
			}
			if(	king_ray_attack(EDGE_RIGHT|EDGE_DOWN, 64-9, k, all_p, is_op&is_b) ) {
				return true;
			}
			if(	king_ray_attack(EDGE_LEFT|EDGE_DOWN, 64-7, k, all_p, is_op&is_b) ) {
				return true;
			}
			if(	king_ray_attack(EDGE_LEFT|EDGE_UP, 9, k, all_p, is_op&is_b) ) {
				return true;
			}
			if(	king_ray_attack(EDGE_RIGHT|EDGE_UP, 7, k, all_p, is_op&is_b) ) {
				return true;
			}
			if(	king_ray_attack(EDGE_LEFT, 1, k, all_p, is_op&is_r) ) {
				return true;
			}
			if(	king_ray_attack(EDGE_RIGHT, 64-1, k, all_p, is_op&is_r) ) {
				return true;
			}
			if(	king_ray_attack(EDGE_UP, 8, k, all_p, is_op&is_r) ) {
				return true;
			}
			return king_ray_attack(EDGE_DOWN, 64-8, k, all_p, is_op&is_r);
		} // kingChecked

		bool __fastcall king_ray_attack_with_discovereds(
					   const uint64 DIR_EDGE,
					   const int SHIFT,
					   const uint64 k,
					   const uint64 all_p,
					   const uint64 is_op,
					   uint64 &att_p,
					   uint64 &att_disc)
		{
			uint64 d;
			d = k;
			while( !(d & (DIR_EDGE)) ) {
				d = ROR(d, SHIFT);
				att_p |= d;
				if( d & all_p ) {
					if( d & is_op ) {
						return true;
					}
					uint64 d2;
					d2 = d;
					while( !(d2 & (DIR_EDGE)) ) {
						d2 = ROR(d2, SHIFT);
						if( d2 & all_p ) {
							if( d2 & is_op ) {
								att_disc |= d;
							}
							return false;
						}
					}
					return false;
				}
			}
			att_p |= d;
			return false;
		}

		bool kingCheckSquares(
			BB_ARGUMENTS,
			uint64 &att_p,
			uint64 &att_b,
			uint64 &att_r,
			uint64 &att_k,
			uint64 &att_disc)
		{
			uint64 k;
			k = k_pos&is_op;
			att_p = 0;
			(void) is_my;
			if( !(k&EDGE_RIGHT) ) {
				if( is_my & (k>>7) & is_pawn )
				{
					return true;
				}
				att_p |= (k>>7);
			}
			
			if( !(k&EDGE_LEFT) ) {
				if( is_my & (k>>9) & is_pawn )
				{
					return true;
				}
				att_p |= (k>>9);
			}
			att_k = knight_moves[LS1B_INDEX(k)];
			if( att_k & is_my & (~(is_pawn|is_b|is_r|k_pos)) ) {
				return true;
			}
			att_b = 0;
			att_r = 0;
			att_disc = 0;
			if(	king_ray_attack_with_discovereds(EDGE_RIGHT|EDGE_DOWN, 64-9, k, all_p, is_my&is_b, att_b, att_disc) ) {
				return true;
			}
			if(	king_ray_attack_with_discovereds(EDGE_LEFT|EDGE_DOWN, 64-7, k, all_p, is_my&is_b, att_b, att_disc) ) {
				return true;
			}
			if(	king_ray_attack_with_discovereds(EDGE_LEFT|EDGE_UP, 9, k, all_p, is_my&is_b, att_b, att_disc) ) {
				return true;
			}
			if(	king_ray_attack_with_discovereds(EDGE_RIGHT|EDGE_UP, 7, k, all_p, is_my&is_b, att_b, att_disc) ) {
				return true;
			}
			if(	king_ray_attack_with_discovereds(EDGE_LEFT, 1, k, all_p, is_my&is_r, att_r, att_disc) ) {
				return true;
			}
			if(	king_ray_attack_with_discovereds(EDGE_RIGHT, 64-1, k, all_p, is_my&is_r, att_r, att_disc) ) {
				return true;
			}
			if(	king_ray_attack_with_discovereds(EDGE_UP, 8, k, all_p, is_my&is_r, att_r, att_disc) ) {
				return true;
			}
			if( king_ray_attack_with_discovereds(EDGE_DOWN, 64-8, k, all_p, is_my&is_r, att_r, att_disc) ) {
				return true;
			}
			return false;
		} // kingCheckSquares

		void makePromoteMoveFlipped(REFBB_ARGUMENTS,
									 uint64 m,									 
									 int &destPieceIndex,
									 state_castleinfo::CastleInfo &castleInfo,
									 Side turn)
		{
			using namespace zobrist;
			uint64 s, d;
			(void) k_pos;
			s = FLIPVERTICAL(m&OPPONENT_LINES) & is_my;
			d = FLIPVERTICAL(m&OPPONENT_LINES) & (~is_my);
			if( d == SQ_A1 ) {
				castleInfo = state_castleinfo::UNSET_OP_CASTLE_LEFT(castleInfo);
			}
			if( d == SQ_H1 ) {
				castleInfo = state_castleinfo::UNSET_OP_CASTLE_RIGHT(castleInfo);
			}
			switch( m & M_PROMOTE_ALL ) {
				case M_PROMOTE_Q:
					if( all_p & d ) {
						if( is_b & d ) {
							if( is_r & d ) {
								destPieceIndex = whiteQueenIndex;
							}
							else {
								is_r |= d;
								destPieceIndex = whiteBishopIndex;
							}
						}
						else {
							if( is_r & d ) {
								destPieceIndex = whiteRookIndex;
								is_b |= d;
							}
							else {
								is_b |= d;
								is_r |= d;
								if( is_pawn & d ) {
									is_pawn &= ~d;
									destPieceIndex = whitePawnIndex;
								}
								else {
									destPieceIndex = whiteKnightIndex;
								}
							}
						}
					}
					else {
						destPieceIndex = noPieceIndex;
						is_b |= d;
						is_r |= d;
					}
					break;
				case M_PROMOTE_R:
					if( all_p & d ) {
						if( is_r & d ) {
							if( is_b & d ) {
								destPieceIndex = whiteQueenIndex;
								is_b &= ~d;
							}
							else {
								destPieceIndex = whiteRookIndex;
							}
						}
						else {
							is_r |= d;
							if( is_b & d ) {
								destPieceIndex = whiteBishopIndex;
								is_b &= ~d;
							}
							else {
								if( is_pawn & d ) {
									is_pawn &= ~d;
									destPieceIndex = whitePawnIndex;
								}
								else {
									destPieceIndex = whiteKnightIndex;
								}
							}
						}
					}
					else {
						destPieceIndex = noPieceIndex;
						is_r |= d;
					}
					break;
				case M_PROMOTE_B:
					if( all_p & d ) {
						if( is_b & d ) {
							if( is_r & d ) {
								destPieceIndex = whiteQueenIndex;
								is_r &= ~d;
							}
							else {
								destPieceIndex = whiteBishopIndex;
							}
						}
						else {
							is_b |= d;
							if( is_r & d ) {
								destPieceIndex = whiteRookIndex;
								is_r &= ~d;
							}
							else {
								if( is_pawn & d ) {
									is_pawn &= ~d;
									destPieceIndex = whitePawnIndex;
								}
								else {
									destPieceIndex = whiteKnightIndex;
								}
							}
						}
					}
					else {
						destPieceIndex = noPieceIndex;
						is_b |= d;
					}
					break;
				case M_PROMOTE_K:
					if( all_p & d ) {
						if( is_b & d ) {
							is_b &= ~d;
							if( is_r & d ) {
								destPieceIndex = whiteQueenIndex;
								is_r &= ~d;
							}
							else {
								destPieceIndex = whiteBishopIndex;
							}
						}
						else {
							if( is_r & d ) {
								destPieceIndex = whiteRookIndex;
								is_r &= ~d;
							}
							else {
								if( is_pawn & d ) {
									is_pawn &= ~d;
									destPieceIndex = whitePawnIndex;
								}
								else {
									destPieceIndex = whiteKnightIndex;
								}
							}
						}
					}
					else {
						destPieceIndex = noPieceIndex;
					}
					break;
			}
			if( turn == BLACK ) {
				++destPieceIndex;
			}
			is_pawn &= ~s;
			is_my ^= s|d;
			is_op &= ~d;
			all_p &= ~s;
			all_p |= d;
		} // makePromoteMoveFlipped

		void makeMoveFlipped(REFBB_ARGUMENTS,
							 uint64 m,
							 int &sourcePieceIndex,
							 int &destPieceIndex,
							 state_castleinfo::CastleInfo &castleInfo,
							 Side turn,
							 int &ep)
		{
			using namespace state_castleinfo;
			using namespace zobrist;

			assert( (king_moves[LS1B_INDEX(is_my&k_pos)] & is_op & k_pos) == 0);

			ep = 0;
			uint64 s, d;
			if( m == CL_SQUARES ) {
				all_p ^= CL_SQUARES_FLIPPED;
				is_my ^= CL_SQUARES_FLIPPED;
				is_r ^= SQ_A8|SQ_D8;
				k_pos ^= SQ_E8|SQ_C8;
				castleInfo = UNSET_MY_CASTLE_BOTH(castleInfo);
				assert( (king_moves[LS1B_INDEX(is_my&k_pos)] & is_op & k_pos) == 0);
				return;
			}
			if( m == CR_SQUARES ) {
				all_p ^= CR_SQUARES_FLIPPED;
				is_my ^= CR_SQUARES_FLIPPED;
				is_r ^= SQ_H8|SQ_F8;
				k_pos ^= SQ_E8|SQ_G8;
				castleInfo = UNSET_MY_CASTLE_BOTH(castleInfo);
				assert( (king_moves[LS1B_INDEX(is_my&k_pos)] & is_op & k_pos) == 0);
				return;
			}
			if( (m & M_PROMOTE_ANY) == M_PROMOTE_ANY) {
				if( turn == WHITE ) {
					sourcePieceIndex = whitePawnIndex;
				}
				else {
					sourcePieceIndex = blackPawnIndex;
				}
				makePromoteMoveFlipped(BB_ARGUMENTS_CALL,
									   m,
									   destPieceIndex,
									   castleInfo,
									   turn);
				assert( (king_moves[LS1B_INDEX(is_my&k_pos)] & is_op & k_pos) == 0);
				return;
			}
			if( (m & M_ENP) == M_ENP ) {
				uint64 t;
				t = FLIPVERTICAL(m&~(M_ENP));
				is_pawn ^= t;
				is_my ^= t;
				all_p ^= t;
				t = (t&is_my) << 8;
				is_pawn ^= t;
				is_op ^= t;
				all_p ^= t;
				if( turn == WHITE ) {
					sourcePieceIndex = whitePawnIndex;
				}
				else {
					sourcePieceIndex = blackPawnIndex;
				}
				destPieceIndex = zobrist::noPieceIndex;
				assert( (king_moves[LS1B_INDEX(is_my&k_pos)] & is_op & k_pos) == 0);
				return;
			}
			m = FLIPVERTICAL(m);
			assert( POPCOUNT(is_my & m) == 1 );
			assert( POPCOUNT(is_op & m) < 2 );
			s = m&is_my;
			d = m-s;
			if( d&all_p ) {
				if( d&is_pawn ) {
					destPieceIndex = whitePawnIndex;
					is_pawn ^= d;
				}
				else {
					if( d&is_b ) {
						is_b ^= d;
						if( d&is_r ) {
							is_r ^= d;
							destPieceIndex = whiteQueenIndex;
						}
						else {
							destPieceIndex = whiteBishopIndex;
						}
					}
					else {
						if( d&is_r ) {
							is_r ^= d;
							destPieceIndex = whiteRookIndex;
						}
						else {
							destPieceIndex = whiteKnightIndex;
						}
					}
				}
			}
			else {
				destPieceIndex = noPieceIndex;
			}
			if( d == SQ_A1 ) {
				castleInfo = UNSET_OP_CASTLE_LEFT(castleInfo);
			}
			if( d == SQ_H1 ) {
				castleInfo = UNSET_OP_CASTLE_RIGHT(castleInfo);
			}
			if( s&is_pawn ) {
				sourcePieceIndex = whitePawnIndex;
				is_pawn ^= m;
				if( d == s>>16 ) {
					ep = LS1B_INDEX(d);
				}
			}
			else {
				if( s&is_b ) {
					is_b ^= m;
					if( s&is_r ) {
						sourcePieceIndex = whiteQueenIndex;
						is_r ^= m;
					}
					else {
						sourcePieceIndex = whiteBishopIndex;
					}
				}
				else {
					if( s&is_r ) {
						sourcePieceIndex = whiteRookIndex;
						is_r ^= m;
						if( s == SQ_A8 ) {
							castleInfo = UNSET_MY_CASTLE_LEFT(castleInfo);
						}
						if( s == SQ_H8 ) {
							castleInfo = UNSET_MY_CASTLE_RIGHT(castleInfo);
						}
					}
					else {
						if( s&k_pos ) {
							sourcePieceIndex = whiteKingIndex;
							k_pos ^= m;
							castleInfo = UNSET_MY_CASTLE_BOTH(castleInfo);
						}
						else {
							sourcePieceIndex = whiteKnightIndex;
						}
					}
				}
			}
			all_p &= ~s;
			all_p |= d;
			is_my ^= m;
			is_op &= ~d;
			if( turn == BLACK ) {
				++sourcePieceIndex;
			}
			else {
				++destPieceIndex;
			}
			assert(k_pos & is_my);
			assert(k_pos & is_op);
			assert( (king_moves[LS1B_INDEX(is_my&k_pos)] & is_op & k_pos) == 0);
		} // makeMoveFlipped

		void genCastleLeft(BB_ARGUMENTS, uint64 *move) {
			(void) is_my;
			if( (all_p&CL_BLOCKS)
				|| ( ( /*LS1B(CL_RA0&all_p) | */LS1B(CL_RA1&all_p) | LS1B(CL_RA2&all_p) ) & is_op & is_r )
				|| ( ( /*LS1B(CL_BA0&all_p) | */LS1B(CL_BA1&all_p) ) & is_op & is_b )
				|| ( ( LS1B(CL_BA2&all_p) | LS1B(CL_BA3&all_p) ) & is_op & is_b )
				|| ( ( LS1B(CL_BA4&all_p) | LS1B(CL_BA5&all_p) ) & is_op & is_b )
				|| ( CL_KNA & is_op & (~(is_r|is_b|k_pos|is_pawn)) ) 
				|| ( CL_PA & is_op & is_pawn )
				|| ( CL_KB & k_pos ) )
			{
				return;
			}
			while( (*move) != 0 ) {
				++move;
			}
			(*move) = CL_SQUARES;
			++move;
			(*move) = 0;
		} // genCastleLeft

		void genCastleRight(BB_ARGUMENTS, uint64 *move) {
			(void) is_my;
			if( (all_p&CR_BLOCKS)
				|| ( ( /*LS1B(CR_RA0&all_p) | */LS1B(CR_RA1&all_p) | LS1B(CR_RA2&all_p) ) & is_op & is_r )
				|| ( ( /*LS1B(CR_BA0&all_p) | */LS1B(CR_BA1&all_p) ) & is_op & is_b )
				|| ( ( LS1B(CR_BA2&all_p) | LS1B(CR_BA3&all_p) ) & is_op & is_b )
				|| ( ( LS1B(CR_BA4&all_p) | LS1B(CR_BA5&all_p) ) & is_op & is_b )
				|| ( CR_KNA & is_op & (~(is_r|is_b|k_pos|is_pawn)) ) 
				|| ( CR_PA & is_op & is_pawn )
				|| ( CR_KB & k_pos ) )
			{
				return;
			}
			while( (*move) != 0 ) {
				++move;
			}
			(*move) = CR_SQUARES;
			++move;
			(*move) = 0;
		} // genCastleRight

		void genEnp(uint64 is_my, uint64 is_pawn, int ep, uint64 *list) {
			assert( ep != 0 );
			while( (*list) != 0 ) {
				++list;
			}
			uint64 ep64 = U64(1)<<ep;
			assert( ep64 & is_pawn & (~is_my) );
			if( ((ep64 & EDGE_LEFT) == 0) && ((ep64>>1) & is_my & is_pawn) ) {
				(*list) = M_ENP | (ep64>>1) | (ep64<<8);
				++list;
			}
			if( ((ep64 & EDGE_RIGHT) == 0) && ((ep64<<1) & is_my & is_pawn) ) {
				(*list) = M_ENP | (ep64<<1) | (ep64<<8);
				++list;
			}
			(*list) = 0;
		}


	} // bitboardmoves
} // zochova
