
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#pragma once

#include "State.h"

#include "include_boost_bind_signals.h"

#include <sstream>
#include <string>
#include <vector>



namespace zochova
{

	namespace log
	{
		class IWriter;
	}

	namespace engine
	{
		namespace search
		{
			class Result;
		}
	}

	namespace communicate
	{
		/*
		template <class T>
		inline std::string to_string (const T& t)
		{
			std::stringstream ss;
			ss << t;
			return ss.str();
		}

		template <class T>
		bool from_string(T& t, 
						 const std::string& s, 
						 std::ios_base& (*f)(std::ios_base&))
		{
		  std::istringstream iss(s);
		  return !(iss >> f >> t).fail();
		}
*/
		void stringSplit(std::string str,
						 std::string delim,
						 std::vector<std::string>& results);

		std::string printBitBoard(uint64 bb);
		bool readMove(Move &m, std::string str);
		char printSquare(Square p);

		class ICommunication
		{
		public:
			static ICommunication* Factory();
			void setLogWriter(log::IWriter* logw);
			virtual void sendOutput(std::string str) = 0;

			boost::signal< void (Move &m) > eventMakeMove;
			boost::signal< void () > eventResign;
			boost::signal< void () > eventOfferDraw;
			boost::signal< void () > eventAcceptDraw;
			boost::signal< void () > eventNewGame;
			boost::signal< void () > eventQuit;
			boost::signal< void () > eventUndo;
			boost::signal< void () > eventRedo;
			boost::signal< void () > eventForce;
			boost::signal< void () > eventMoveNow;
			boost::signal< void () > eventSetEngineBlack;
			boost::signal< void () > eventSetEngineWhite;
			boost::signal< void (int moves, int t1, int t2) > eventSetTimeControl;
			boost::signal< void (State* st) > eventNeedCurrentState;
			boost::signal< void () > eventSetPonder;
			boost::signal< void () > eventUnSetPonder;
			boost::signal< void (int t) > eventSetUserTime;
			boost::signal< void (int t) > eventSetEngineTime;

			virtual void makeMove(const Move &m) =0;
			virtual void offerDraw() = 0;
			virtual void acceptDraw() = 0;
			virtual void moveAccepted() = 0;
			virtual void moveIllegal() = 0;
			virtual void resign() = 0;
			virtual void newThinkingLine(const engine::search::Result& result) = 0;
			virtual void setUsersTurn() {
				isUsersTurn = true;
			};
			virtual void unsetUsersTurn() {
				isUsersTurn = false;
			};
			virtual void run();
		protected:
			State currentState;
			bool isUsersTurn;
			bool processed;
			void doInputThread();
			virtual bool newInput(std::string command) = 0;
			log::IWriter* logger;
			std::string logInputPrefix;
			bool isShowThinking;
		}; // ICommunication

	} // communicate
} // zochova
