
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "evaluation.h"
#include "bit_boards.h"

#include "search_flags.h"

namespace zochova
{
	namespace engine
	{
		namespace evaluation
		{

			std::ostream &operator<<(std::ostream &stream, Value val) {
				stream << (int)val;
				return stream;
			}
			
			int pawnKingStructure(uint64 pawns, uint64 k_pos) {
				int r, x;
				uint64 pos, t;
				if( k_pos&CASTLED_KING_POS ) {
					r = CASTLED_KING_VALUE;
				}
				else {
					if( k_pos&KING_HIDING_POS ) {
						r = HIDING_KING_VALUE;
					}
					else {
						r = 0;
					}
				}
				if( k_pos&SQ_A1 ) {
					if( (k_pos<<8) & pawns ) {
						r += HIDING_KING_VALUE;
					}
					if( (k_pos<<9) & pawns ) {
						r += HIDING_KING_VALUE;
					}
				}
				else {
					if( k_pos&SQ_H1 ) {
						if( (k_pos<<8) & pawns ) {
							r += HIDING_KING_VALUE;
						}
						if( (k_pos<<7) & pawns ) {
							r += HIDING_KING_VALUE;
						}
					}
					else {
						r += POPCOUNT( SECOND_ROW & pawns & ((k_pos<<8) | (k_pos<<9) | (k_pos<<7)) ) * HIDING_KING_VALUE;

					}

				}
			//	r += POPCOUNT( SECOND_ROW & pawns & ((k_pos<<8) | (k_pos<<9) | (k_pos<<7)) ) * HIDING_KING_VALUE;
				t = pawns;
				do {
					pos = LS1B(t);
					r += PAWN_VALUE;
					if( pawns & (pos<<9) ) {
						r += PAWN_CHAIN_VALUE;
					}
					if( pawns & (pos<<7) ) {
						r += PAWN_CHAIN_VALUE;
					}
				} while( (t -= pos) != 0 );
				pos = EDGE_RIGHT;
				for(;;) {
					if( (x = POPCOUNT( pawns & pos )) > 1 ) {
						r += x * DOUBLE_PAWN_VALUE;
					}
					if( pos == EDGE_LEFT ) {
						return r;
					}
					pos >>= 1;
				};
			} // pawnKingStructure

			int rookStructure(uint64 rooks, uint64 is_my, uint64 is_pawn) {
				int r;
				uint64 rook;
				r = 0;
				do {
					uint64 file;
					r += ROOK_VALUE;
					rook = LS1B(rooks);
					if( rook & SQ_A1 ) {
						if( is_pawn & is_my & (SQ_A2|SQ_A3|SQ_A4) ) {
							r += ROOK_CORNER_CLOSED_VALUE;
						}
						if( SQ_B1 & is_my ) {
							r += ROOK_CORNER_CLOSED_VALUE;
						}
					}
					else {
						if( rook & SQ_H1 ) {
							if( is_pawn & is_my & (SQ_H2|SQ_H3|SQ_H4) ) {
								r += ROOK_CORNER_CLOSED_VALUE;
							}
							if( SQ_G1 & is_my ) {
								r += ROOK_CORNER_CLOSED_VALUE;
							}
						}
						else {
							if( rook & SQ_B1 ) {
								if( is_pawn & is_my & (SQ_B2|SQ_B3|SQ_B4) ) {
									r += ROOK_CORNER_CLOSED_VALUE;
								}
								if( SQ_C1 & is_my ) {
									r += ROOK_CORNER_CLOSED_VALUE;
								}
							}
							else {
								if( rook & SQ_G1 ) {
									if( is_pawn & is_my & (SQ_G2|SQ_G3|SQ_G4) ) {
										r += ROOK_CORNER_CLOSED_VALUE;
									}
									if( SQ_F1 & is_my ) {
										r += ROOK_CORNER_CLOSED_VALUE;
									}
								}
								else {
									if( rook & SQ_G2 ) {
										if( is_pawn & is_my & (SQ_G3|SQ_G4) ) {
											r += ROOK_CORNER_CLOSED_VALUE;
										}
									}
									else {
										if( rook & SQ_A2 ) {
											if( is_pawn & is_my & (SQ_A3|SQ_A4) ) {
												r += ROOK_CORNER_CLOSED_VALUE;
											}
										}
										else {
											if( rook & OPPONENT_LINES ) {
												r += ROOK_AT_OPPONENT_VALUE;
											}
										}
									}
								}
							}
						}
					}
					file = EDGE_LEFT << (LS1B_INDEX(rook)%8);
					if( (file & rooks) - rook ) {
						r += ROOKS_ON_SAME_FILE_VALUE;
					}
					file &= ~(rook|(rook-1));
					if( !(file & is_my) ) {
						r += ROOK_ON_OPEN_FILE_VALUE;
					}
					else {
						if( !(file & is_my & is_pawn) ) {
							r += ROOK_ON_SEMI_OPEN_FILE_VALUE;
						}
					}
				} while( rooks -= rook );
				return r;
			} // rookStructure

			int bishopStructure(uint64 bishops, uint64 myPawns, uint64 opPawns) {
				int r;
				uint64 bishop;
				r = 0;
				do {
					uint64 color;
					r += BISHOP_VALUE;
					bishop = LS1B(bishops);
					if( bishop & BISHOP_PENALTY_SQ_1 ) {
						r += BISHOP_PENALTY_1_VALUE;
					}
					else {
						if( bishop & BISHOP_PENALTY_SQ_2 ) {
							r += BISHOP_PENALTY_2_VALUE;
						}
					}
					if( SQ_COLOR_0 & bishop ) {
						color = SQ_COLOR_0;
					}
					else {
						color = SQ_COLOR_1;
					}
					if( (POPCOUNT(myPawns&color) + (2 * POPCOUNT(opPawns&color))) > 9 ) {
						r += BISHOP_PAWN_BLOCKED_PENALTY_VALUE * ((POPCOUNT(myPawns&color) + (2 * POPCOUNT(opPawns&color))) / 5);
					}
				} while( bishops -= bishop );
				return r;
			} // bishopStructure
/*
			int staticValue(BB_ARGUMENTS) {
				int r;
				r = 0;
				if( (POPCOUNT(is_my&(~is_pawn)) < 4) || (POPCOUNT(is_op&(~is_pawn)) < 4) ) {
					r += POPCOUNT(is_my&is_pawn&ROW_5) * PAWN_RANK5_VALUE;
					r += POPCOUNT(is_my&is_pawn&ROW_6) * PAWN_RANK6_VALUE;
					r += POPCOUNT(is_my&is_pawn&ROW_7) * PAWN_RANK7_VALUE;
					r -= POPCOUNT(is_op&is_pawn&ROW_4) * PAWN_RANK5_VALUE;
					r -= POPCOUNT(is_op&is_pawn&ROW_3) * PAWN_RANK6_VALUE;
					r -= POPCOUNT(is_op&is_pawn&ROW_2) * PAWN_RANK7_VALUE;
				}
				else {
					r = (POPCOUNT(is_my&(~k_pos)&CENTER_SQ) - POPCOUNT(is_op&(~k_pos)&CENTER_SQ)) * CENTER_SQ_VALUE;
					r += (POPCOUNT(is_my&(~k_pos)&NEAR_CENTER_SQ) - POPCOUNT(is_op&(~k_pos)&NEAR_CENTER_SQ)) * NEAR_CENTER_SQ_VALUE;
					if( (k_pos & is_my & SQ_G1) && !(is_r & is_my & SQ_H1) ) {
						r += CASTLED_KING_VALUE;
					}
					else {
						if( (k_pos & is_my & SQ_C1) && !(is_r & is_my & SQ_A1) ) {
							r += CASTLED_KING_VALUE;
						}
					}
					if( (k_pos & is_op & SQ_G8) && !(is_r & is_op & SQ_H8) ) {
						r -= CASTLED_KING_VALUE;
					}
					else {
						if( (k_pos & is_op & SQ_C8) && !(is_r & is_op & SQ_A8) ) {
							r -= CASTLED_KING_VALUE;
						}
					}
				}

				r += (POPCOUNT(is_pawn&is_my) - POPCOUNT(is_pawn&is_op)) * PAWN_VALUE;
				r += (POPCOUNT(is_r&is_my) - POPCOUNT(is_r&is_op)) * ROOK_VALUE;
				r += (POPCOUNT(is_b&is_my) - POPCOUNT(is_b&is_op)) * BISHOP_VALUE;
				r += (POPCOUNT(is_r&is_b&is_my) - POPCOUNT(is_r&is_b&is_op)) * (QUEEN_VALUE - BISHOP_VALUE - ROOK_VALUE);
				r += (POPCOUNT(is_my&~(is_b|is_r|is_pawn|k_pos)) - POPCOUNT(is_op&~(is_b|is_r|is_pawn|k_pos))) * KNIGHT_VALUE;

//				r += (POPCOUNT(is_r&is_b&is_my) - POPCOUNT(is_r&is_b&is_op)) * (QUEEN_VALUE);
				assert( r < evaluation::INFINITE_VALUE );
				assert( r > -evaluation::INFINITE_VALUE );

				return r;
			}
*/
			int staticValue(const State &state) {
				BB_DECLARE;
				zochova::state_castleinfo::CastleInfo ci;
				state.genBB(BB_ARGUMENTS_CALL, ci);
				return staticValue(BB_ARGUMENTS_CALL);
			}

			int staticValue(BB_ARGUMENTS) {
				(void) all_p;
				int r;
				r = 0;
			//	#ifdef EVAL_STATIC
			//	if( (POPCOUNT(is_my&(~is_pawn)) < 4) || (POPCOUNT(is_op&(~is_pawn)) < 4) ) {
			//	#else
				if( false ) {
			//	#endif
					//
					//  endgame

					r += (POPCOUNT(is_my&CENTER_SQ&(~is_pawn)&(~k_pos)) - POPCOUNT(is_op&CENTER_SQ&(~is_pawn)&(~k_pos))) * CENTER_SQ_VALUE;
					r += POPCOUNT(is_my&is_pawn&ROW_5) * PAWN_RANK5_VALUE;
					r += POPCOUNT(is_my&is_pawn&ROW_6) * PAWN_RANK6_VALUE;
					r += POPCOUNT(is_my&is_pawn&ROW_7) * PAWN_RANK7_VALUE;
					r -= POPCOUNT(is_op&is_pawn&ROW_4) * PAWN_RANK5_VALUE;
					r -= POPCOUNT(is_op&is_pawn&ROW_3) * PAWN_RANK6_VALUE;
					r -= POPCOUNT(is_op&is_pawn&ROW_2) * PAWN_RANK7_VALUE;
					r += (POPCOUNT(is_pawn&is_my) - POPCOUNT(is_pawn&is_op)) * PAWN_VALUE;
					r += (POPCOUNT(is_r&is_my) - POPCOUNT(is_r&is_op)) * ROOK_VALUE;
					r += (POPCOUNT(is_b&is_my) - POPCOUNT(is_b&is_op)) * BISHOP_VALUE;
					r += (POPCOUNT(is_r&is_b&is_my) - POPCOUNT(is_r&is_b&is_op)) * (QUEEN_VALUE - BISHOP_VALUE - ROOK_VALUE);
					r += (POPCOUNT(is_my&~(is_b|is_r|is_pawn|k_pos)) - POPCOUNT(is_op&~(is_b|is_r|is_pawn|k_pos))) * KNIGHT_VALUE;
				}
				else {
					//
					//  middlegame or opening
					r += (POPCOUNT(is_my&CENTER_SQ&(~k_pos)) - POPCOUNT(is_op&CENTER_SQ&(~k_pos))) * CENTER_SQ_VALUE;
					r += (POPCOUNT(is_my&NEAR_CENTER_SQ&(~is_pawn)&(~k_pos)) - POPCOUNT(is_op&NEAR_CENTER_SQ&(~is_pawn)&(~k_pos))) * NEAR_CENTER_SQ_VALUE;
					if( is_pawn&is_my ) {
						r += pawnKingStructure(is_pawn&is_my, k_pos&is_my);
					}
					if( is_pawn&is_op ) {
						r -= pawnKingStructure(FLIPVERTICAL(is_pawn&is_op), FLIPVERTICAL(k_pos&is_op));
					}


				/*	if( is_r&is_my&(~is_b) ) {
						r += rookStructure(is_r&is_my&(~is_b), is_my, is_pawn);
					}
					if( is_r&is_op&(~is_b) ) {
						r -= rookStructure(FLIPVERTICAL(is_r&is_op&(~is_b)), FLIPVERTICAL(is_my), FLIPVERTICAL(is_pawn));
					}
					if( is_b&is_my&(~is_r) ) {
						r += bishopStructure(is_b&is_my&(~is_r), is_pawn&is_my, is_pawn&is_op);
					}
					if( is_b&is_op&(~is_r) ) {
						r -= bishopStructure(FLIPVERTICAL(is_b&is_op&(~is_r)), FLIPVERTICAL(is_pawn&is_op), FLIPVERTICAL(is_pawn&is_my));
					}
				*/

//					r += (POPCOUNT(is_b&is_my) - POPCOUNT(is_b&is_op)) * BISHOP_VALUE;

				//	r += (POPCOUNT(is_pawn&is_my) - POPCOUNT(is_pawn&is_op)) * PAWN_VALUE;
					r += (POPCOUNT(is_r&is_my) - POPCOUNT(is_r&is_op)) * ROOK_VALUE;
					r += (POPCOUNT(is_b&is_my) - POPCOUNT(is_b&is_op)) * BISHOP_VALUE;
					r += (POPCOUNT(is_r&is_b&is_my) - POPCOUNT(is_r&is_b&is_op)) * (QUEEN_VALUE - BISHOP_VALUE - ROOK_VALUE);
					
				//	r += (POPCOUNT(is_r&is_b&is_my) - POPCOUNT(is_r&is_b&is_op)) * QUEEN_VALUE;
					
					r += (POPCOUNT(is_my&~(is_b|is_r|is_pawn|k_pos)) - POPCOUNT(is_op&~(is_b|is_r|is_pawn|k_pos))) * KNIGHT_VALUE;

					r += (bitboardmoves::mobility(BB_ARGUMENTS_CALL) / 2);
					r -= (bitboardmoves::mobility(
						FLIPVERTICAL(all_p), 
						FLIPVERTICAL(is_op), 
						FLIPVERTICAL(is_my), 
						FLIPVERTICAL(is_pawn),
						FLIPVERTICAL(is_b), 
						FLIPVERTICAL(is_r), 
						FLIPVERTICAL(k_pos)	) / 2);

				}
				return r;
			} // staticValue

		} // evaluation
	} // engine
} // zochova
