
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#pragma once

#include "communicate.h"

#include "include_boost_lexical_cast.h"

#include <iostream>
#include <fstream>
#include <string>
#include <stack>


namespace zochova
{
	namespace log
	{

		class IWriter
		{
		public:
			IWriter();

			virtual ~IWriter() {};

			void pushWriteTimeFlag();

			void popWriteTimeFlag();

			void setWriteTime();

			void unsetWriteTime();

		protected:

			std::stack< bool > writeTimeFlagStack;

			virtual IWriter& write(std::string str) = 0;

			bool isOpen;

			bool WriteTimeFlag;

			template<typename T> friend IWriter& operator<<(IWriter& w, T &obj);
		}; // IWriter

		class EmtpyWriter : public IWriter
		{
			IWriter& write(std::string str) { (void) str; return *this; };
		};

		class RealWriter: public IWriter
		{
		public:
			RealWriter();
			~RealWriter();
			RealWriter(std::string argFileName, bool append = false);
		private:
			IWriter& write(std::string str);
			std::ofstream fileStream;
			std::string fileName;
			bool afterNewLine;
		}; // RealWriter

		template<typename T> inline IWriter& operator<<(IWriter& w, T &obj) {
			if( w.isOpen ) {
				return w.write(boost::lexical_cast<std::string>(obj));
			}
			else {
				return w;
			}
		}
	} // log

} // zochova
