
/***************************************************************************

  (\=,                                                               ,=/)
 //  .\                                                             /.  \\ 
(( \_  \                                                           /  _/ ))
 ))  `\_) ZZZZZZZ   OOO     CCCCC H     H   OOO   V     V    A    (_/'  (( 
(/     \       Z   O   O   C      H     H  O   O  V     V   A A    /     \)
 | _.-'|      Z   O     O C       H     H O     O V     V  A   A   |'-._ | 
  )___(      Z    O     O C       HHHHHHH O     O V     V AAAAAAA   )___(  
 (=====)    Z     O     O C       H     H O     O  V   V  A     A  (=====) 
 }====={   Z       O   O   C      H     H  O   O    V V   A     A  }====={ 
(_______) ZZZZZZZ   OOO     CCCCC H     H   OOO      V    A     A (_______)

***************************************************************************/

/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "Connections.h"
#include "log.h"
#include "OpeningBooks.h"

#include <iostream>

#ifdef _MSC_VER
//
//  This is here to disable some warnings, 
//  because MS Visual C++ static code analyser 
//  yields lots of warnings in boost code
#pragma warning(push, 3)
#pragma warning(push)
#pragma warning(disable : ALL_CODE_ANALYSIS_WARNINGS )
#endif

#include "boost/program_options.hpp"
#include "boost/thread.hpp"

#ifdef _MSC_VER
#pragma warning(pop)
#pragma warning(pop)
#endif


boost::program_options::variables_map* handle_options(int argc, char *argv[]) {
	using namespace boost::program_options;

	options_description opts("Zochova options");
	opts.add_options()(
		"booktype",
		value< std::string >(),
		"type of opening book to use");
	opts.add_options()(
		"rebelmain",
		value< std::string >(),
		"rebel main book");
	opts.add_options()(
		"rebeltour",
		value< std::string >(),
		"rebel tournament book");
	opts.add_options()(
		"rebelrandom1",
		value< std::string >(),
		"rebel random1 book");
	opts.add_options()(
		"rebelrandom2",
		value< std::string >(),
		"rebel random2 book");
	opts.add_options()(
		"logcomm",
		value< std::string >(),
		"turn on IO logging, into the given file ");
	opts.add_options()(
		"logcomma",
		"IO log file appending ( default is overwriting )");
	opts.add_options()(
		"logengine",
		value< std::string >(),
		"turn on engine logging, into the given file ");
	opts.add_options()(
		"logenginea",
		"engine log file appending ( default is overwriting )");
	opts.add_options()(
		"threads",
		value< int >(),
		"search threads (defaults to number cpu cores detected)");
	opts.add_options()(
		"help",
		"print help on command line options");		
	variables_map *vm = new variables_map();
	try {
		store( command_line_parser(argc, argv).options(opts).run(), *vm );
	} catch( std::exception& e ) {
		std::cout << e.what() << "\n";
		exit(1);
	}
	notify(*vm);
	if( vm->count("help") > 0 ) {
		std::cout << opts << std::endl;
		exit(1);
	}
	return vm;
}

int main(int argc, char *argv[]) {
//	_CrtSetBreakAlloc(7472);

	using namespace std;
	using namespace zochova;
	boost::program_options::variables_map *varmap = handle_options(argc, argv);
    cout.rdbuf()->pubsetbuf(NULL,0);
	cin.rdbuf()->pubsetbuf(NULL,0);


	//
	//  Create IO log
	log::IWriter* commLog;
	if( (*varmap).count("logcomm") > 0 ) {
		try {
			commLog = new log::RealWriter(
				(*varmap)["logcomm"].as<string>(),
				(*varmap).count("logcomma") > 0	);
		}
		catch(...) {
			cout << "Can't write IO log to file: " << (*varmap)["logcomm"].as<string>() << endl;
			commLog = new log::EmtpyWriter();
		}
	}
	else {
		commLog = new log::EmtpyWriter();
	}

	//
	//  Create engine log
	log::IWriter* engineLog;
	if( (*varmap).count("logengine") > 0 ) {
		try {
			engineLog = new log::RealWriter(
				(*varmap)["logengine"].as<string>(),
				(*varmap).count("logenginea") > 0	);
		}
		catch(...) {
			cout << "Can't write engine log to file: " << (*varmap)["logengine"].as<string>() << endl;
			engineLog = new log::EmtpyWriter();
		}
	}
	else {
		engineLog = new log::EmtpyWriter();
	}

	//
	//  Choose number of threads to use
	int nThreads;
	if( (*varmap).count("threads") > 0 ) {
		try {
			nThreads = (*varmap)["threads"].as<int>();
			if( (nThreads < 1) || (nThreads > 64) ) {
				cout << "thread counts 1 to 64 are supported, defaulting to detectable core count" << endl;
				throw;
			}
		}
		catch(...) {
			nThreads = boost::thread::hardware_concurrency();
		}
	}
	else {
		nThreads = boost::thread::hardware_concurrency();
	}
	if( nThreads == 0 ) {
		cout << "Unknown core count, using 1 thread";
		nThreads = 1;
	}

	//  While using multiple trheads is not implemented, defaulting to 1
	nThreads = 1;

	//
	//  Choose the opening book to use
	IOpeningBook* openingBook;
	try {
		if( (*varmap)["booktype"].as< std::string>() == "rebel" ) {
			if( (*varmap).count("rebeltour") == 1 ) {
				openingBook = new RebelBook(
					(*varmap)["rebelmain"].as< std::string>(),
					(*varmap)["rebelrandom1"].as< std::string>(),
					(*varmap)["rebelrandom2"].as< std::string>(),
					(*varmap)["rebeltour"].as< std::string>() );
			}
			else {
				openingBook = new RebelBook(
					(*varmap)["rebelmain"].as< std::string>(),
					(*varmap)["rebelrandom1"].as< std::string>(),
					(*varmap)["rebelrandom2"].as< std::string>(),
					"" );
			}
		}
		else if( (*varmap)["booktype"].as< std::string>() == "pgsql" ) {
			// TODO
			openingBook = new EmptyOpeningBook();
		}
		else {
			openingBook = new EmptyOpeningBook();
		}
	}
	catch( ... ) {
		openingBook = new EmptyOpeningBook();
	}

	delete varmap;

	//_CrtDumpMemoryLeaks();

	//
	//  Start the application
	Connections::run(commLog, engineLog, nThreads, openingBook);

	//_CrtDumpMemoryLeaks();

	//
	//  Clean up
	delete engineLog;
	delete commLog;
	delete openingBook;

	_CrtDumpMemoryLeaks();

	return 1;
}
